# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mplang/protos/v1alpha1/mpir.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!mplang/protos/v1alpha1/mpir.proto\x12\x16mplang.protos.v1alpha1\"\xde\x03\n\tAttrProto\x12>\n\x04type\x18\x01 \x01(\x0e\x32*.mplang.protos.v1alpha1.AttrProto.AttrTypeR\x04type\x12\x0c\n\x01\x66\x18\x02 \x01(\x02R\x01\x66\x12\x0c\n\x01i\x18\x03 \x01(\x03R\x01i\x12\x0c\n\x01s\x18\x04 \x01(\tR\x01s\x12\x0c\n\x01\x62\x18\x05 \x01(\x08R\x01\x62\x12\x1b\n\traw_bytes\x18\x06 \x01(\x0cR\x08rawBytes\x12\x16\n\x06\x66loats\x18\x07 \x03(\x02R\x06\x66loats\x12\x12\n\x04ints\x18\x08 \x03(\x03R\x04ints\x12\x12\n\x04strs\x18\t \x03(\tR\x04strs\x12\x35\n\x04\x66unc\x18\n \x01(\x0b\x32!.mplang.protos.v1alpha1.FuncProtoR\x04\x66unc\x12\x38\n\x05graph\x18\x0b \x01(\x0b\x32\".mplang.protos.v1alpha1.GraphProtoR\x05graph\"\x8a\x01\n\x08\x41ttrType\x12\r\n\tUNDEFINED\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\x07\n\x03INT\x10\x02\x12\n\n\x06STRING\x10\x03\x12\x08\n\x04\x42OOL\x10\x04\x12\t\n\x05\x42YTES\x10\x05\x12\n\n\x06\x46LOATS\x10\x06\x12\x08\n\x04INTS\x10\x07\x12\x0b\n\x07STRINGS\x10\x08\x12\x0c\n\x08\x46UNCTION\x10\n\x12\t\n\x05GRAPH\x10\x0b\"\x87\x02\n\tFuncProto\x12\x12\n\x04type\x18\x01 \x01(\tR\x04type\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x12\n\x04\x62ody\x18\x03 \x01(\tR\x04\x62ody\x12\x1d\n\ndoc_string\x18\x04 \x01(\tR\tdocString\x12\x42\n\x05\x61ttrs\x18\x05 \x03(\x0b\x32,.mplang.protos.v1alpha1.FuncProto.AttrsEntryR\x05\x61ttrs\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01\"h\n\x0fTensorTypeProto\x12\x36\n\x05\x64type\x18\x01 \x01(\x0e\x32 .mplang.protos.v1alpha1.DataTypeR\x05\x64type\x12\x1d\n\nshape_dims\x18\x02 \x03(\x03R\tshapeDims\"\xaf\x01\n\x0eTableTypeProto\x12G\n\x07\x63olumns\x18\x01 \x03(\x0b\x32-.mplang.protos.v1alpha1.TableTypeProto.ColumnR\x07\x63olumns\x1aT\n\x06\x43olumn\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x36\n\x05\x64type\x18\x02 \x01(\x0e\x32 .mplang.protos.v1alpha1.DataTypeR\x05\x64type\"\xe8\x02\n\x0bMPTypeProto\x12J\n\x0btensor_type\x18\x01 \x01(\x0b\x32\'.mplang.protos.v1alpha1.TensorTypeProtoH\x00R\ntensorType\x12G\n\ntable_type\x18\x02 \x01(\x0b\x32&.mplang.protos.v1alpha1.TableTypeProtoH\x00R\ttableType\x12\x14\n\x05pmask\x18\x03 \x01(\x03R\x05pmask\x12\x44\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32..mplang.protos.v1alpha1.MPTypeProto.AttrsEntryR\x05\x61ttrs\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01\x42\x0b\n\ttype_spec\"\xd2\x02\n\tNodeProto\x12\x17\n\x07op_type\x18\x01 \x01(\tR\x06opType\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x16\n\x06inputs\x18\x03 \x03(\tR\x06inputs\x12@\n\touts_info\x18\x05 \x03(\x0b\x32#.mplang.protos.v1alpha1.MPTypeProtoR\x08outsInfo\x12\x42\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32,.mplang.protos.v1alpha1.NodeProto.AttrsEntryR\x05\x61ttrs\x12\x1d\n\ndoc_string\x18\x06 \x01(\tR\tdocString\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01\"e\n\x0bVersionInfo\x12\x14\n\x05major\x18\x01 \x01(\x05R\x05major\x12\x14\n\x05minor\x18\x02 \x01(\x05R\x05minor\x12\x14\n\x05patch\x18\x03 \x01(\x05R\x05patch\x12\x14\n\x05\x65xtra\x18\x04 \x01(\tR\x05\x65xtra\"\xc0\x02\n\nGraphProto\x12=\n\x07version\x18\x01 \x01(\x0b\x32#.mplang.protos.v1alpha1.VersionInfoR\x07version\x12\x37\n\x05nodes\x18\x02 \x03(\x0b\x32!.mplang.protos.v1alpha1.NodeProtoR\x05nodes\x12\x18\n\x07outputs\x18\x03 \x03(\tR\x07outputs\x12\x43\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32-.mplang.protos.v1alpha1.GraphProto.AttrsEntryR\x05\x61ttrs\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01*\xea\x01\n\x08\x44\x61taType\x12\r\n\tUNDEFINED\x10\x00\x12\x06\n\x02U8\x10\x01\x12\x06\n\x02I8\x10\x02\x12\x07\n\x03U16\x10\x03\x12\x07\n\x03I16\x10\x04\x12\x07\n\x03U32\x10\x05\x12\x07\n\x03I32\x10\x06\x12\x07\n\x03U64\x10\x07\x12\x07\n\x03I64\x10\x08\x12\x07\n\x03\x46\x31\x36\x10\t\x12\x07\n\x03\x46\x33\x32\x10\n\x12\x07\n\x03\x46\x36\x34\x10\x0b\x12\n\n\x06STRING\x10\x0c\x12\x08\n\x04\x42OOL\x10\r\x12\r\n\tCOMPLEX64\x10\x0e\x12\x0e\n\nCOMPLEX128\x10\x0f\x12\x0c\n\x08\x42\x46LOAT16\x10\x10\x12\x08\n\x04\x44\x41TE\x10\x11\x12\x08\n\x04TIME\x10\x12\x12\r\n\tTIMESTAMP\x10\x13\x12\x08\n\x04JSON\x10\x14\x42.Z,github.com/secretflow/mplang/protos/v1alpha1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mplang.protos.v1alpha1.mpir_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z,github.com/secretflow/mplang/protos/v1alpha1'
  _globals['_FUNCPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_FUNCPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_MPTYPEPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_MPTYPEPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_NODEPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_NODEPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_GRAPHPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_DATATYPE']._serialized_start=2223
  _globals['_DATATYPE']._serialized_end=2457
  _globals['_ATTRPROTO']._serialized_start=62
  _globals['_ATTRPROTO']._serialized_end=540
  _globals['_ATTRPROTO_ATTRTYPE']._serialized_start=402
  _globals['_ATTRPROTO_ATTRTYPE']._serialized_end=540
  _globals['_FUNCPROTO']._serialized_start=543
  _globals['_FUNCPROTO']._serialized_end=806
  _globals['_FUNCPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_FUNCPROTO_ATTRSENTRY']._serialized_end=806
  _globals['_TENSORTYPEPROTO']._serialized_start=808
  _globals['_TENSORTYPEPROTO']._serialized_end=912
  _globals['_TABLETYPEPROTO']._serialized_start=915
  _globals['_TABLETYPEPROTO']._serialized_end=1090
  _globals['_TABLETYPEPROTO_COLUMN']._serialized_start=1006
  _globals['_TABLETYPEPROTO_COLUMN']._serialized_end=1090
  _globals['_MPTYPEPROTO']._serialized_start=1093
  _globals['_MPTYPEPROTO']._serialized_end=1453
  _globals['_MPTYPEPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_MPTYPEPROTO_ATTRSENTRY']._serialized_end=806
  _globals['_NODEPROTO']._serialized_start=1456
  _globals['_NODEPROTO']._serialized_end=1794
  _globals['_NODEPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_NODEPROTO_ATTRSENTRY']._serialized_end=806
  _globals['_VERSIONINFO']._serialized_start=1796
  _globals['_VERSIONINFO']._serialized_end=1897
  _globals['_GRAPHPROTO']._serialized_start=1900
  _globals['_GRAPHPROTO']._serialized_end=2220
  _globals['_GRAPHPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_GRAPHPROTO_ATTRSENTRY']._serialized_end=806
# @@protoc_insertion_point(module_scope)



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from mplang.core.dtype import DType
from mplang.core.mask import Mask
from mplang.core.mptype import MPType
from mplang.core.table import TableType
from mplang.core.tensor import Shape

if TYPE_CHECKING:
    from mplang.core.cluster import ClusterSpec


class MPContext:
    """The context of an MPObject.

    MPContext is the abstract base class for all execution contexts.
    It only holds the immutable cluster_spec plus lightweight parent/root
    helpers used to support stack-scoped extension state (attached lazily by
    external features on the root context).
    """

    def __init__(self, cluster_spec: ClusterSpec, *, parent: MPContext | None = None):
        if cluster_spec is None:
            raise ValueError("cluster_spec cannot be None")
        self.cluster_spec = cluster_spec
        # Parent link enables stack-scoped state sharing: ephemeral child contexts
        # (e.g. short-lived tracing) can delegate to a stable root without relying
        # on process-wide globals.
        self._parent: MPContext | None = parent

    # Basic topology helpers
    def world_size(self) -> int:
        return len(self.cluster_spec.nodes)

    @property
    def parent(self) -> MPContext | None:
        """Direct parent context or None if this is root."""
        return self._parent

    def root(self) -> MPContext:
        """Return the root context (follow parent chain)."""
        ctx: MPContext = self
        visited: set[int] = set()
        while ctx._parent is not None:
            if id(ctx) in visited:
                raise RuntimeError("Cycle detected in MPContext parent chain")
            visited.add(id(ctx))
            ctx = ctx._parent
        return ctx


class MPObject(ABC):
    """The base class for all objects in mp-system."""

    @property
    @abstractmethod
    def mptype(self) -> MPType:
        """The type information of the object.

        This property is readonly (mandatory) and will be used for JAX compilation
        to determine the appropriate data type during trace and compilation phases.
        MPType can be passed between different MPObjects as a value.
        """

    @property
    def dtype(self) -> DType:
        return self.mptype.dtype

    @property
    def shape(self) -> Shape:
        return self.mptype.shape

    @property
    def schema(self) -> TableType:
        """The table schema of the object.

        Only available for table types.
        """
        return self.mptype.schema

    @property
    def pmask(self) -> Mask | None:
        return self.mptype.pmask

    @property
    def attrs(self) -> dict[str, Any]:
        return self.mptype.attrs

    @property
    @abstractmethod
    def ctx(self) -> MPContext:
        """Return the context of the object."""


# Forward docstrings from MPType to MPObject
MPObject.dtype.__doc__ = MPType.dtype.__doc__
MPObject.shape.__doc__ = MPType.shape.__doc__
MPObject.schema.__doc__ = MPType.schema.__doc__
MPObject.pmask.__doc__ = MPType.pmask.__doc__
MPObject.attrs.__doc__ = MPType.attrs.__doc__



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from collections.abc import Callable
from enum import Enum
from functools import wraps
from typing import Any

from jax.tree_util import tree_unflatten

from mplang.core import Mask, MPObject, Rank, peval, psize
from mplang.core.context_mgr import cur_ctx
from mplang.ops import spu
from mplang.simp import mpi


class SecureAPI(ABC):
    """Base class for secure APIs."""

    @abstractmethod
    def seal(self, obj: MPObject, frm_mask: Mask | None) -> list[MPObject]: ...

    @abstractmethod
    def sealFrom(self, obj: MPObject, root: Rank) -> MPObject: ...

    @abstractmethod
    def seval(self, fe_type: str, pyfn: Callable, *args: Any, **kwargs: Any) -> Any:
        """Run a function in the secure environment."""

    @abstractmethod
    def reveal(self, obj: MPObject, to_mask: Mask) -> MPObject: ...

    @abstractmethod
    def revealTo(self, obj: MPObject, to_rank: Rank) -> MPObject: ...


class Delegation(SecureAPI):
    """Delegate to a trusted third-party to perform secure operations."""

    def seal(self, obj: MPObject, frm_mask: Mask | None = None) -> list[MPObject]:
        raise NotImplementedError("TODO")

    def sealFrom(self, obj: MPObject, root: Rank) -> MPObject:
        raise NotImplementedError("TODO")

    def seval(self, fe_type: str, pyfn: Callable, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError("TODO")

    def reveal(self, obj: MPObject, to_mask: Mask) -> MPObject:
        raise NotImplementedError("TODO")

    def revealTo(self, obj: MPObject, to_rank: Rank) -> MPObject:
        raise NotImplementedError("TODO")


class SPU(SecureAPI):
    """Use SPU to  perform secure operations."""

    def get_spu_mask(self) -> Mask:
        spu_devices = cur_ctx().cluster_spec.get_devices_by_kind("SPU")
        if not spu_devices:
            raise ValueError("No SPU device found in the cluster specification")
        if len(spu_devices) > 1:
            raise ValueError("Multiple SPU devices found in the cluster specification")
        spu_device = spu_devices[0]
        spu_mask = Mask.from_ranks([member.rank for member in spu_device.members])
        return spu_mask

    def seal(self, obj: MPObject, frm_mask: Mask | None = None) -> list[MPObject]:
        spu_mask: Mask = self.get_spu_mask()
        if obj.pmask is None:
            if frm_mask is None:
                # NOTE: The length of the return list is statically determined by obj_mask,
                # so only static masks are supported here.
                raise ValueError("Seal does not support dynamic masks.")
            else:
                # Force seal from the given mask, the runtime will raise error if the mask
                # does not match obj.pmask.
                # TODO(jint): add set_pmask primitive.
                pass
        else:
            if frm_mask is None:
                frm_mask = obj.pmask
            else:
                if not Mask(frm_mask).is_subset(obj.pmask):
                    raise ValueError(f"Cannot seal from {frm_mask} to {obj.pmask}, ")

        # Get the world_size from spu_mask (number of parties in SPU computation)
        world_size = Mask(spu_mask).num_parties()
        pfunc, ins, _ = spu.makeshares(
            obj, world_size=world_size, visibility=spu.Visibility.SECRET
        )
        assert len(ins) == 1
        shares = peval(pfunc, ins, frm_mask)

        # scatter the shares to each party.
        return [mpi.scatter_m(spu_mask, rank, shares) for rank in Mask(frm_mask)]

    def sealFrom(self, obj: MPObject, root: Rank) -> MPObject:
        results = seal(obj, frm_mask=Mask.from_ranks(root))
        assert len(results) == 1, f"Expected one result, got {len(results)}"
        return results[0]

    def seval(self, fe_type: str, pyfn: Callable, *args: Any, **kwargs: Any) -> Any:
        if fe_type != "jax":
            raise ValueError(f"Unsupported fe_type: {fe_type}")

        spu_mask = self.get_spu_mask()
        pfunc, in_vars, out_tree = spu.jax_compile(pyfn, *args, **kwargs)
        assert all(var.pmask == spu_mask for var in in_vars), in_vars
        out_flat = peval(pfunc, in_vars, spu_mask)
        return tree_unflatten(out_tree, out_flat)

    def reveal(self, obj: MPObject, to_mask: Mask) -> MPObject:
        spu_mask = self.get_spu_mask()

        assert obj.pmask == spu_mask, (obj.pmask, spu_mask)

        # (n_parties, n_shares)
        shares = [mpi.bcast_m(to_mask, rank, obj) for rank in Mask(spu_mask)]
        assert len(shares) == Mask(spu_mask).num_parties(), (shares, spu_mask)
        assert all(share.pmask == to_mask for share in shares)

        # Reconstruct the original object from shares
        pfunc, ins, _ = spu.reconstruct(*shares)
        return peval(pfunc, ins, to_mask)[0]  # type: ignore[no-any-return]

    def revealTo(self, obj: MPObject, to_rank: Rank) -> MPObject:
        return self.reveal(obj, to_mask=Mask.from_ranks(to_rank))


class SEE(Enum):
    """Secure Execution Environment."""

    MOCK = 0
    SPU = 1
    TEE = 2


# TODO(jint): move me to options.py
mode: SEE = SEE.SPU


def _get_sapi() -> SecureAPI:
    """Get the current secure API based on the mode."""
    if mode == SEE.MOCK:
        return Delegation()
    elif mode == SEE.SPU:
        return SPU()
    elif mode == SEE.TEE:
        raise NotImplementedError("TEE is not implemented yet")
    else:
        raise ValueError(f"Unknown mode: {mode}")


# seal :: m a -> [s a]
def seal(obj: MPObject, frm_mask: Mask | None = None) -> list[MPObject]:
    """Seal an simp object, result a list of sealed objects, with
    the i'th element as the secret from the i'th party.
    """
    return _get_sapi().seal(obj, frm_mask=frm_mask)


# sealFrom :: m a -> m Rank -> s a
def sealFrom(obj: MPObject, root: Rank) -> MPObject:
    """Seal an simp object from a specific root party."""
    return _get_sapi().sealFrom(obj, root)


# reveal :: s a -> m a
def reveal(obj: MPObject, to_mask: Mask | None = None) -> MPObject:
    """Reveal a sealed object to pmask'ed parties."""
    to_mask = to_mask or Mask.all(psize())
    return _get_sapi().reveal(obj, to_mask)


# revealTo :: s a -> m Rank -> m a
def revealTo(obj: MPObject, to_rank: Rank) -> MPObject:
    return _get_sapi().revealTo(obj, to_rank)


# srun :: (a -> a) -> s a -> s a
def srun(pyfn: Callable, *, fe_type: str = "jax") -> Callable:
    @wraps(pyfn)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        return _get_sapi().seval(fe_type, pyfn, *args, **kwargs)

    return wrapped



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from mplang.core.dtype import UINT8, DType
from mplang.core.pfunc import PFunction
from mplang.core.table import TableLike, TableType
from mplang.core.tensor import TensorLike, TensorType
from mplang.kernels import base
from mplang.kernels.base import KernelContext, get_kernel_spec, kernel_exists

# Default bindings
# Import kernel implementation modules explicitly so their @kernel_def entries
# register at import time. Keep imports grouped; alias with leading underscore
# to silence unused variable warnings without F401 pragmas.
_IMPL_IMPORTED = False


def _ensure_impl_imported() -> None:
    global _IMPL_IMPORTED
    if _IMPL_IMPORTED:
        return
    from mplang.kernels import basic as _impl_basic  # noqa: F401
    from mplang.kernels import crypto as _impl_crypto  # noqa: F401
    from mplang.kernels import mock_tee as _impl_tee  # noqa: F401
    from mplang.kernels import phe as _impl_phe  # noqa: F401
    from mplang.kernels import spu as _impl_spu  # noqa: F401
    from mplang.kernels import sql_duckdb as _impl_sql_duckdb  # noqa: F401
    from mplang.kernels import stablehlo as _impl_stablehlo  # noqa: F401

    _IMPL_IMPORTED = True


# imports consolidated above

_DEFAULT_BINDINGS: dict[str, str] = {
    # basic
    "basic.identity": "basic.identity",
    "basic.read": "basic.read",
    "basic.write": "basic.write",
    "basic.constant": "basic.constant",
    "basic.rank": "basic.rank",
    "basic.prand": "basic.prand",
    "basic.table_to_tensor": "basic.table_to_tensor",
    "basic.tensor_to_table": "basic.tensor_to_table",
    "basic.debug_print": "basic.debug_print",
    "basic.pack": "basic.pack",
    "basic.unpack": "basic.unpack",
    # crypto
    "crypto.keygen": "crypto.keygen",
    "crypto.enc": "crypto.enc",
    "crypto.dec": "crypto.dec",
    "crypto.kem_keygen": "crypto.kem_keygen",
    "crypto.kem_derive": "crypto.kem_derive",
    "crypto.hkdf": "crypto.hkdf",
    # phe
    "phe.keygen": "phe.keygen",
    "phe.encrypt": "phe.encrypt",
    "phe.mul": "phe.mul",
    "phe.add": "phe.add",
    "phe.decrypt": "phe.decrypt",
    "phe.dot": "phe.dot",
    "phe.gather": "phe.gather",
    "phe.scatter": "phe.scatter",
    "phe.concat": "phe.concat",
    "phe.reshape": "phe.reshape",
    "phe.transpose": "phe.transpose",
    # spu
    "spu.seed_env": "spu.seed_env",
    "spu.makeshares": "spu.makeshares",
    "spu.reconstruct": "spu.reconstruct",
    "spu.run_pphlo": "spu.run_pphlo",
    # stablehlo
    "mlir.stablehlo": "mlir.stablehlo",
    # sql
    # generic SQL op; backend-specific kernel id for duckdb
    "sql.run": "duckdb.run_sql",
    # tee
    # "tee.quote_gen": "mock_tee.quote_gen",
    # "tee.attest": "mock_tee.attest",
}


# --- RuntimeContext ---


class RuntimeContext:
    """Per-runtime execution context with isolated op->kernel bindings.

    This object owns ONLY static dispatch metadata ("op bindings") and mutable
    per-rank kernel side state/cache/stats. It does NOT store per-evaluation
    variable bindings (those are provided to the evaluator at evaluation time).

    Parameters
    ----------
    rank : int
        Local rank of this participant.
    world_size : int
        Total number of participants.
    initial_bindings : Mapping[str, str] | None, optional
        Optional partial overrides applied on top of the default binding table
        during construction (override semantics, not replace). These map
        op_type -> kernel_id and form a *template* for dispatch. After
        initialization, all (re)binding must go through ``bind_op`` /
        ``rebind_op`` on this context (scoped to THIS runtime only).
    state : dict, optional
        Mutable per-runtime key/value storage for kernels. Flat key space;
        callers SHOULD use dotted prefixes (e.g. "stablehlo.compile_cache").
        Kernels own their *state* (functional correctness data, caches,
        handles, compiled objects, RNGs, etc.). Runtime does not interpret
        structure—values may themselves be dicts if a kernel wants its own
        pocket. Created empty when omitted.
    stats : dict, optional
        Mutable statistics/telemetry owned by the runtime (usage counters,
        timings, profiling aids). Kernels may increment counters but should
        avoid storing functional state here. A default "op_calls" mapping is
        ensured. Created empty when omitted.
    """

    __slots__ = ("_ibindings", "rank", "state", "stats", "world_size")

    def __init__(
        self,
        rank: int,
        world_size: int,
        initial_bindings: Mapping[str, str] | None = None,
        *,
        state: dict[str, Any] | None = None,
        stats: dict[str, Any] | None = None,
    ) -> None:
        _ensure_impl_imported()
        self.rank = rank
        self.world_size = world_size
        # Merge defaults with user overrides (override semantics)
        self._ibindings: dict[str, str] = {
            **_DEFAULT_BINDINGS,
            **(initial_bindings or {}),
        }
        self.state = state if state is not None else {}
        self.stats = stats if stats is not None else {}
        self.stats.setdefault("op_calls", {})

    def run_kernel(self, pfunc: PFunction, arg_list: list[Any]) -> list[Any]:
        fn_type = pfunc.fn_type
        kid = self._ibindings.get(fn_type)
        if kid is None:
            raise NotImplementedError(f"no backend kernel registered for op {fn_type}")
        spec = get_kernel_spec(kid)
        fn = spec.fn  # kernel implementation
        if len(arg_list) != len(pfunc.ins_info):
            raise ValueError(
                f"kernel {fn_type} arg count mismatch: got {len(arg_list)}, expect {len(pfunc.ins_info)}"
            )
        for idx, (ins_spec, val) in enumerate(
            zip(pfunc.ins_info, arg_list, strict=True)
        ):
            if isinstance(ins_spec, TableType):
                _validate_table_arg(fn_type, idx, ins_spec, val)
                continue
            if isinstance(ins_spec, TensorType):
                _validate_tensor_arg(fn_type, idx, ins_spec, val)
                continue

        # install kernel context
        kctx = KernelContext(rank=self.rank, world_size=self.world_size, runtime=self)
        token = base._CTX_VAR.set(kctx)
        try:
            raw = fn(pfunc, *arg_list)
        finally:
            base._CTX_VAR.reset(token)

        try:
            op_calls = self.stats.setdefault("op_calls", {})
            op_calls[fn_type] = op_calls.get(fn_type, 0) + 1
        except Exception:  # pragma: no cover - never raise due to stats
            pass
        expected = len(pfunc.outs_info)
        if expected == 0:
            if raw in (None, (), []):
                return []
            raise ValueError(
                f"kernel {fn_type} should return no values; got {type(raw).__name__}"
            )
        if expected == 1:
            if isinstance(raw, (tuple, list)):
                if len(raw) != 1:
                    raise ValueError(
                        f"kernel {fn_type} produced {len(raw)} outputs, expected 1"
                    )
                return [raw[0]]
            return [raw]
        if not isinstance(raw, (tuple, list)):
            raise TypeError(
                f"kernel {fn_type} must return sequence (len={expected}), got {type(raw).__name__}"
            )
        if len(raw) != expected:
            raise ValueError(
                f"kernel {fn_type} produced {len(raw)} outputs, expected {expected}"
            )
        return list(raw)

    def reset(self) -> None:
        self.state.clear()

    # ---- runtime state API (flat key space) ----
    # Keys are treated atomically; convention encourages dotted prefixes
    # (e.g. 'stablehlo.compile_cache.hash', 'crypto.rng'). Implementation
    # does NOT parse or create hierarchical dicts—any grouping is purely
    # by string prefix. Values themselves MAY be dicts if callers want a
    # manual pocket. This keeps semantics simple and predictable.

    def ensure_state(self, key: str, factory: type | Any = dict) -> Any:
        """Return value for key; if absent create via factory and store.

        Key is not parsed; dotted forms are allowed but treated as a single
        map key. Use consistent prefixes for grouping (e.g. 'spu.config').
        """
        if not key:
            raise ValueError("empty state key")
        val = self.state.get(key)
        if val is None:
            val = factory()
            self.state[key] = val
        return val

    def get_state(self, key: str, default: Any | None = None) -> Any:
        if not key:
            raise ValueError("empty state key")
        return self.state.get(key, default)

    def set_state(self, key: str, value: Any) -> None:
        if not key:
            raise ValueError("empty state key")
        self.state[key] = value

    def del_state(self, key: str) -> None:
        if not key:
            raise ValueError("empty state key")
        self.state.pop(key, None)

    def list_state(self, prefix: str = "") -> dict[str, Any]:
        """Return mapping of key -> value; optional prefix filter.

        Prefix match is string-based; if prefix is non-empty include keys
        where key == prefix or key starts with prefix + '.'.
        """
        if not prefix:
            return dict(self.state)
        pref = prefix if prefix.endswith(".") else prefix + "."
        out: dict[str, Any] = {}
        for k, v in self.state.items():
            if k == prefix or k.startswith(pref):
                out[k] = v
        return out

    # ---- explicit (re)binding API ----
    def bind_op(self, op_type: str, kernel_id: str, *, force: bool = False) -> None:
        """Bind an operation to a kernel for THIS context only.

        force=False (default) keeps existing binding (no silent override).
        """
        if not kernel_exists(kernel_id):
            raise KeyError(f"kernel_id {kernel_id} not registered")
        if not force and op_type in self._ibindings:
            return
        self._ibindings[op_type] = kernel_id

    def rebind_op(self, op_type: str, kernel_id: str) -> None:
        """Force rebind an operation to a different kernel (shorthand)."""
        self.bind_op(op_type, kernel_id, force=True)

    # Introspection helpers
    def list_bound_ops(self) -> list[str]:  # pragma: no cover - convenience
        return sorted(self._ibindings.keys())

    def get_binding(self, op_type: str) -> str | None:  # pragma: no cover
        return self._ibindings.get(op_type)

    def __repr__(self) -> str:  # pragma: no cover - debug aid
        return (
            f"RuntimeContext(rank={self.rank}, world_size={self.world_size}, "
            f"bound_ops={len(self._ibindings)})"
        )


def _validate_table_arg(
    fn_type: str, arg_index: int, spec: TableType, value: Any
) -> None:
    if not isinstance(value, TableLike):
        raise TypeError(
            f"kernel {fn_type} input[{arg_index}] expects TableLike, got {type(value).__name__}"
        )
    if len(value.columns) != len(spec.columns):
        raise ValueError(
            f"kernel {fn_type} input[{arg_index}] column count mismatch: got {len(value.columns)}, expected {len(spec.columns)}"
        )


def _validate_tensor_arg(
    fn_type: str, arg_index: int, spec: TensorType, value: Any
) -> None:
    # Backend-only handle sentinel (e.g., PHE keys) bypasses all structural checks
    if tuple(spec.shape) == (-1, 0) and spec.dtype == UINT8:
        return

    if isinstance(value, (int, float, bool, complex)):
        val_shape: tuple[Any, ...] = ()
        duck_dtype: Any = type(value)
    else:
        if not isinstance(value, TensorLike):
            raise TypeError(
                f"kernel {fn_type} input[{arg_index}] expects TensorLike, got {type(value).__name__}"
            )
        val_shape = getattr(value, "shape", ())
        duck_dtype = getattr(value, "dtype", None)

    if len(spec.shape) != len(val_shape):
        raise ValueError(
            f"kernel {fn_type} input[{arg_index}] rank mismatch: got {val_shape}, expected {spec.shape}"
        )

    for dim_idx, (spec_dim, val_dim) in enumerate(
        zip(spec.shape, val_shape, strict=True)
    ):
        if spec_dim >= 0 and spec_dim != val_dim:
            raise ValueError(
                f"kernel {fn_type} input[{arg_index}] shape mismatch at dim {dim_idx}: got {val_dim}, expected {spec_dim}"
            )

    try:
        val_dtype = DType.from_any(duck_dtype)
    except (ValueError, TypeError):  # pragma: no cover
        raise TypeError(
            f"kernel {fn_type} input[{arg_index}] has unsupported dtype object {duck_dtype!r}"
        ) from None
    if val_dtype != spec.dtype:
        raise ValueError(
            f"kernel {fn_type} input[{arg_index}] dtype mismatch: got {val_dtype}, expected {spec.dtype}"
        )



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SPU flat backend kernel tests.

Rewritten to remove deprecated SpuHandler usage. These tests exercise the
registered kernels:
    - spu.makeshares
    - spu.reconstruct
    - spu.run_pphlo (compiled via frontend.spu.jax_compile)

We keep the tests intentionally lean: just enough coverage to ensure the new
flat kernel pathway functions for single-party (REF2K) and multi-party (SEMI2K)
execution flows.
"""

from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor

import jax.numpy as jnp
import numpy as np
import spu.libspu as libspu

from mplang.core.cluster import ClusterSpec, Device, Node, RuntimeInfo
from mplang.core.dtype import DType
from mplang.core.mpobject import MPContext, MPObject
from mplang.core.mptype import MPType
from mplang.core.pfunc import PFunction
from mplang.core.tensor import TensorType
from mplang.kernels.base import list_kernels
from mplang.kernels.context import RuntimeContext
from mplang.kernels.spu import SpuValue
from mplang.kernels.value import TensorValue
from mplang.ops import spu
from mplang.runtime.link_comm import LinkCommunicator


class DummyContext(MPContext):
    """Minimal MPContext for compiling JAX functions to SPU PFunction."""

    def __init__(self):
        runtime = RuntimeInfo(version="dev", platform="local", op_bindings={})
        node = Node(name="p0", rank=0, endpoint="local", runtime_info=runtime)
        device = Device(name="p0_local", kind="local", members=[node])
        spec = ClusterSpec(nodes={node.name: node}, devices={device.name: device})
        super().__init__(spec)


class DummyTensor(MPObject):
    """Minimal MPObject used purely for spu.jax_compile shape/dtype inference."""

    def __init__(self, shape: tuple[int, ...], dtype: jnp.dtype):  # type: ignore[valid-type]
        self._mptype = MPType.tensor(DType.from_any(dtype), shape)
        self._ctx = DummyContext()

    @property
    def mptype(self) -> MPType:  # pragma: no cover - trivial
        return self._mptype

    @property
    def ctx(self) -> MPContext:  # pragma: no cover - trivial
        return self._ctx


def create_mem_link_contexts(world_size: int) -> list[LinkCommunicator]:
    """Create memory link communicators for each party.

    Returns a list so tests can still iterate, but initialize_spu_runtime will
    now be called per-rank with a single link instance.
    """
    addrs = [f"P{i}" for i in range(world_size)]
    return [LinkCommunicator(rank, addrs, mem_link=True) for rank in range(world_size)]


def _compile_add(shape: tuple[int, ...] = (3,), dtype=jnp.float32):  # type: ignore[valid-type]
    def add_fn(x, y):
        return x + y

    args = [DummyTensor(shape, dtype), DummyTensor(shape, dtype)]
    pfunc, _ins, _out_tree = spu.jax_compile(add_fn, *args)
    return pfunc


def _makeshares_pfunc(arr: np.ndarray, world_size: int) -> PFunction:
    """Create makeshares PFunction with outs matching expected party world size.

    We rely on caller providing world_size consistent with runtime creation config.
    """
    tensor_type = TensorType.from_obj(arr)
    outs = tuple(tensor_type for _ in range(world_size))
    return PFunction(
        fn_type="spu.makeshares",
        ins_info=(tensor_type,),
        outs_info=outs,
        visibility=libspu.Visibility.VIS_SECRET.value,
    )


def _reconstruct_pfunc(example: np.ndarray, world_size: int) -> PFunction:
    """Create a reconstruct PFunction whose ins_info matches world size.

    Each input is a share with same tensor type meta; backend kernel validates
    actual share objects are SpuValue.
    """
    tensor_type = TensorType.from_obj(example)
    return PFunction(
        fn_type="spu.reconstruct",
        ins_info=tuple(tensor_type for _ in range(world_size)),
        outs_info=(tensor_type,),
    )


class TestSpuKernels:
    def test_kernel_registry(self):
        for name in ["spu.makeshares", "spu.reconstruct", "spu.run_pphlo"]:
            assert name in list_kernels()

    def test_makeshares_reconstruct_single_party(self):
        world = 1
        cfg = libspu.RuntimeConfig(
            protocol=libspu.ProtocolKind.REF2K, field=libspu.FieldType.FM128
        )
        link_ctxs = create_mem_link_contexts(world)
        runtime = RuntimeContext(rank=0, world_size=world)
        # Seed SPU env via kernel
        seed_fn = PFunction(
            fn_type="spu.seed_env",
            ins_info=(),
            outs_info=(),
            config=cfg,
            world=world,
            link=link_ctxs[0],
        )
        runtime.run_kernel(seed_fn, [])

        x = np.array([1, 2, 3], dtype=np.int32)
        mk = _makeshares_pfunc(x, world)
        shares = runtime.run_kernel(mk, [TensorValue(x)])
        assert len(shares) == 1 and isinstance(shares[0], SpuValue)
        rc = _reconstruct_pfunc(x, world)
        # run_kernel expects outs length=1; we call reconstruct with the share list
        out = runtime.run_kernel(rc, shares)[0]
        np.testing.assert_array_equal(out.to_numpy(), x)

    def test_makeshares_reconstruct_multiparty(self):
        world = 2
        cfg = libspu.RuntimeConfig(
            protocol=libspu.ProtocolKind.SEMI2K, field=libspu.FieldType.FM128
        )
        link_ctxs = create_mem_link_contexts(world)
        # seed rank0 state
        runtime0 = RuntimeContext(rank=0, world_size=world)
        seed0 = PFunction(
            fn_type="spu.seed_env",
            ins_info=(),
            outs_info=(),
            config=cfg,
            world=world,
            link=link_ctxs[0],
        )
        runtime0.run_kernel(seed0, [])
        # seed rank1 state to simulate multi-rank kernel contexts
        runtime1 = RuntimeContext(rank=1, world_size=world)
        seed1 = PFunction(
            fn_type="spu.seed_env",
            ins_info=(),
            outs_info=(),
            config=cfg,
            world=world,
            link=link_ctxs[1],
        )
        runtime1.run_kernel(seed1, [])

        x = np.array([10, 20], dtype=np.int32)
        mk = _makeshares_pfunc(x, world)
        shares = runtime0.run_kernel(mk, [TensorValue(x)])
        assert len(shares) == world and all(isinstance(s, SpuValue) for s in shares)
        rc = _reconstruct_pfunc(x, world)
        out = runtime0.run_kernel(rc, list(shares))[0]
        np.testing.assert_array_equal(out.to_numpy(), x)

    def test_spu_run_pphlo_single_party(self):
        world = 1
        cfg = libspu.RuntimeConfig(
            protocol=libspu.ProtocolKind.REF2K, field=libspu.FieldType.FM128
        )
        link_ctxs = create_mem_link_contexts(world)
        runtime = RuntimeContext(rank=0, world_size=world)
        seed = PFunction(
            fn_type="spu.seed_env",
            ins_info=(),
            outs_info=(),
            config=cfg,
            world=world,
            link=link_ctxs[0],
        )
        runtime.run_kernel(seed, [])

        pfunc = _compile_add((3,), jnp.float32)
        x = np.array([1.0, 2.0, 3.0], dtype=np.float32)
        y = np.array([4.0, 5.0, 6.0], dtype=np.float32)
        expected = x + y

        mkx = _makeshares_pfunc(x, world)
        mky = _makeshares_pfunc(y, world)
        x_shares = runtime.run_kernel(mkx, [TensorValue(x)])
        y_shares = runtime.run_kernel(mky, [TensorValue(y)])

        # Single-party run (rank=0)
        result_share = runtime.run_kernel(pfunc, [x_shares[0], y_shares[0]])[0]
        assert isinstance(result_share, SpuValue)
        rc = _reconstruct_pfunc(expected, world)
        out = runtime.run_kernel(rc, [result_share])[0]
        np.testing.assert_allclose(out.to_numpy(), expected, rtol=1e-5)

    def test_spu_run_pphlo_multiparty(self):
        world = 2
        cfg = libspu.RuntimeConfig(
            protocol=libspu.ProtocolKind.SEMI2K, field=libspu.FieldType.FM128
        )
        link_ctxs = create_mem_link_contexts(world)

        # Initialize per-rank runtime & seed (store explicit runtimes)
        runtimes = {}
        for r in range(world):
            rt = RuntimeContext(rank=r, world_size=world)
            runtimes[r] = rt
            seed_fn = PFunction(
                fn_type="spu.seed_env",
                ins_info=(),
                outs_info=(),
                config=cfg,
                world=world,
                link=link_ctxs[r],
            )
            rt.run_kernel(seed_fn, [])  # use explicit runtime

        pfunc = _compile_add((3,), jnp.float32)
        x = np.array([1.0, 2.0, 3.0], dtype=np.float32)
        y = np.array([4.0, 5.0, 6.0], dtype=np.float32)
        expected = x + y

        mkx = _makeshares_pfunc(x, world)
        mky = _makeshares_pfunc(y, world)
        x_shares: list[SpuValue] = runtimes[0].run_kernel(mkx, [TensorValue(x)])
        y_shares: list[SpuValue] = runtimes[0].run_kernel(mky, [TensorValue(y)])

        # Run spu.run_pphlo concurrently per rank to satisfy interactive protocol
        def party(rank: int, xs: SpuValue, ys: SpuValue):
            rt = runtimes[rank]
            return rt.run_kernel(pfunc, [xs, ys])[0]

        with ThreadPoolExecutor(max_workers=world) as pool:
            futures = [
                pool.submit(party, r, x_shares[r], y_shares[r]) for r in range(world)
            ]
            results = [f.result() for f in futures]

        assert len(results) == world and all(isinstance(r, SpuValue) for r in results)
        rc = _reconstruct_pfunc(expected, world)
        out = runtimes[0].run_kernel(rc, results)[0]
        np.testing.assert_allclose(out.to_numpy(), expected, rtol=1e-5)



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time
import unittest
from contextlib import contextmanager

import jax.numpy as jnp
import xgboost as xgb
from hist_jax import (
    XGBoostJAX,
    _compute_best_split_per_node,
    build_bins_vmapped,
    build_histogram,
    compute_best_split,
    compute_indices_vmapped,
)
from sklearn.datasets import make_classification, make_regression
from sklearn.metrics import accuracy_score, mean_squared_error, roc_auc_score


def _calculate_gain_naively(G, H, reg_lambda, gamma, min_child_weight):
    """A simple, unoptimized reference implementation to calculate best split."""
    n_features, n_bins = G.shape
    best_gain = -jnp.inf
    best_feat = -1
    best_thresh_idx = -1

    for i in range(n_features):
        G_total = jnp.sum(G[i, :])
        H_total = jnp.sum(H[i, :])

        if H_total < min_child_weight:
            continue

        score_parent = G_total**2 / (H_total + reg_lambda)

        G_left, H_left = 0.0, 0.0
        for j in range(n_bins - 1):  # Iterate through all possible split points
            G_left += G[i, j]
            H_left += H[i, j]

            if H_left < min_child_weight:
                continue

            G_right = G_total - G_left
            H_right = H_total - H_left

            if H_right < min_child_weight:
                continue

            score_left = G_left**2 / (H_left + reg_lambda)
            score_right = G_right**2 / (H_right + reg_lambda)

            current_gain = 0.5 * (score_left + score_right - score_parent) - gamma

            if current_gain > best_gain:
                best_gain = current_gain
                best_feat = i
                best_thresh_idx = j

    return best_gain, best_feat, best_thresh_idx


@contextmanager
def suppress_stdout():
    """
    A context manager to temporarily suppress stdout.

    Usage:
        with suppress_stdout():
            print("This will not be seen")
        print("This will be seen")
    """
    devnull_fd = os.open(os.devnull, os.O_WRONLY)
    original_stdout_fd = sys.stdout.fileno()
    saved_stdout_fd = os.dup(original_stdout_fd)

    try:
        os.dup2(devnull_fd, original_stdout_fd)
        yield
    finally:
        os.dup2(saved_stdout_fd, original_stdout_fd)
        os.close(devnull_fd)
        os.close(saved_stdout_fd)


class TestOptimizedXGBoostFunctions(unittest.TestCase):
    """
    Unit tests for the optimized JAX-based XGBoost functions.
    """

    def test_build_bins(self):
        """Tests bin creation for both varying and constant features."""
        x = jnp.array([[1.0, 5.0], [3.0, 5.0], [5.0, 5.0]])
        max_bins = 5
        # The function build_bins_equi_width is vmapped
        bins = build_bins_vmapped(x, max_bins)

        expected_splits_0 = jnp.linspace(1.0, 5.0, num=max_bins + 1)[1:-1]
        expected_splits_1 = jnp.full(
            max_bins - 1, jnp.inf
        )  # Constant features return inf splits

        self.assertEqual(bins.shape, (2, max_bins - 1))
        self.assertTrue(jnp.allclose(bins[0], expected_splits_0))
        self.assertTrue(jnp.allclose(bins[1], expected_splits_1))

    def test_compute_bin_indices(self):
        """Tests the mapping of values to bin indices, including edge cases."""
        x = jnp.array([[0.5, 8.0], [2.5, 5.0], [5.5, 3.0]])
        # Splits for feature 0: 2.0, 3.0, 4.0
        # Splits for feature 1: 4.0, 5.0, 6.0
        bins = jnp.array([[2.0, 3.0, 4.0], [4.0, 5.0, 6.0]])

        bin_indices = compute_indices_vmapped(x, bins)

        # Feat 0: 0.5(<2)->0; 2.5(>2,<=3)->1; 5.5(>4)->3 (indices 0,1,2,3 for 4 bins)
        # Feat 1: 8.0(>6)->3; 5.0(>4,<=5)->1; 3.0(<4)->0
        expected_indices = jnp.array([[0, 3], [1, 1], [3, 0]])
        self.assertTrue(jnp.array_equal(bin_indices, expected_indices))

    def test_build_histogram_optimized(self):
        """Tests the fully vectorized histogram building."""
        g = jnp.array([1.0, 2.0, 3.0, 4.0])
        h = jnp.array([0.1, 0.2, 0.3, 0.4])
        bt_local = jnp.array([0, 0, 1, 1])
        bin_indices = jnp.array([[0, 1], [1, 0], [0, 1], [1, 0]])
        t, k = 2, 2

        GH_hist = build_histogram(g, h, bt_local, bin_indices, t, k)
        self.assertEqual(GH_hist.shape, (2, 2, 2, 2))

        expected_G_hist = jnp.array([
            [[1.0, 2.0], [3.0, 4.0]],
            [[2.0, 1.0], [4.0, 3.0]],
        ])
        expected_H_hist = jnp.array([
            [[0.1, 0.2], [0.3, 0.4]],
            [[0.2, 0.1], [0.4, 0.3]],
        ])

        self.assertTrue(jnp.allclose(GH_hist[..., 0], expected_G_hist))
        self.assertTrue(jnp.allclose(GH_hist[..., 1], expected_H_hist))

    def test_compute_best_split_per_node_optimized(self):
        """Tests the gain calculation for a single node from a combined GH histogram."""
        GH_node = jnp.array([
            [[1.0, 0.5], [2.0, 1.0], [3.0, 1.5]],
            [[4.0, 2.0], [5.0, 2.5], [6.0, 3.0]],
        ])
        reg_lambda, gamma, min_child_weight = 1.0, 0.0, 1.0

        best_gain, best_feat, best_thresh_idx = _compute_best_split_per_node(
            GH_node, reg_lambda, gamma, min_child_weight
        )

        # Get result from the naive "golden standard" function
        G, H = GH_node[..., 0], GH_node[..., 1]
        naive_gain, naive_feat, naive_thresh_idx = _calculate_gain_naively(
            G, H, reg_lambda, gamma, min_child_weight
        )

        self.assertTrue(jnp.allclose(best_gain, naive_gain))
        self.assertEqual(best_feat, naive_feat)
        self.assertEqual(best_thresh_idx, naive_thresh_idx)

    def test_compute_best_split_optimized(self):
        """Tests the vectorized best split computation over multiple nodes."""
        GH_hist = jnp.array([
            [[[1.0, 0.5], [2.0, 1.0], [3.0, 1.5]]],
            [[[10.0, 5.0], [20.0, 10.0], [30.0, 15.0]]],
        ]).transpose((1, 0, 2, 3))

        reg_lambda, gamma, min_child_weight = 1.0, 0.0, 0.0

        max_gains, best_features, best_threshold_idxs = compute_best_split(
            GH_hist, reg_lambda, gamma, min_child_weight
        )

        self.assertEqual(max_gains.shape, (2,))
        self.assertEqual(best_features.shape, (2,))
        self.assertEqual(best_threshold_idxs.shape, (2,))

        _, feat0, idx0 = _compute_best_split_per_node(
            GH_hist[:, 0, :, :], reg_lambda, gamma, min_child_weight
        )
        self.assertEqual(best_features[0], feat0)
        self.assertEqual(best_threshold_idxs[0], idx0)

        _, feat1, idx1 = _compute_best_split_per_node(
            GH_hist[:, 1, :, :], reg_lambda, gamma, min_child_weight
        )
        self.assertEqual(best_features[1], feat1)
        self.assertEqual(best_threshold_idxs[1], idx1)

    def test_classification(self):
        """Encapsulated test for the classification task."""
        print("\n" + "=" * 50 + "\n--- Running Classification Test ---\n" + "=" * 50)
        # Load and preprocess data
        X_np, y_np = make_classification(
            n_samples=50_000,
            n_features=20,
            n_informative=10,
            n_redundant=5,
            n_classes=2,
            random_state=42,
        )
        X = jnp.array(X_np, dtype=jnp.float32)
        y = jnp.array(y_np, dtype=jnp.float32)

        # Model hyperparameters
        params = {
            "n_estimators": 10,
            "learning_rate": 0.1,
            "max_depth": 3,
            "max_bin": 64,
            "reg_lambda": 1.0,
            "gamma": 0.0,
            "min_child_weight": 1.0,
        }

        print("--- 1. Testing JAX Classifier ---")
        model_jax = XGBoostJAX(**params, objective="binary:logistic")

        start_run = time.time()
        model_jax.fit(X, y)
        with suppress_stdout():
            print(model_jax.model)
        print(
            f"JAX Classifier compiled + first fit in: {time.time() - start_run:.4f} seconds"
        )

        start_run = time.time()
        model_jax.fit(X, y)
        with suppress_stdout():
            print(model_jax.model)
        print(f"JAX Classifier execution time: {time.time() - start_run:.4f} seconds")

        start_run = time.time()
        with suppress_stdout():
            print(model_jax.predict(X))
        print(
            f"JAX Classifier prediction compiled time: {time.time() - start_run:.4f} seconds"
        )

        start_run = time.time()
        y_pred_proba = model_jax.predict(X)
        with suppress_stdout():
            print(y_pred_proba)
        print(f"JAX Classifier prediction time: {time.time() - start_run:.4f} seconds")

        accuracy = accuracy_score(y, (y_pred_proba > 0.5).astype(int))
        auc = roc_auc_score(y, y_pred_proba)
        print(f"JAX Model -> Accuracy: {accuracy:.4f}, AUC: {auc:.4f}\n")

        print("--- 2. Testing Official XGBoost for comparison ---")
        xgb_params = params.copy()
        model_xgb = xgb.XGBClassifier(
            **xgb_params,
            tree_method="hist",
            objective="binary:logistic",
            eval_metric="logloss",
        )

        start_run = time.time()
        model_xgb.fit(X_np, y_np)
        print(f"Official XGBoost fit time: {time.time() - start_run:.4f} seconds")
        y_pred_proba_xgb = model_xgb.predict_proba(X_np)[:, 1]
        accuracy_xgb = accuracy_score(y_np, (y_pred_proba_xgb > 0.5).astype(int))
        auc_xgb = roc_auc_score(y_np, y_pred_proba_xgb)
        print(f"Official XGBoost -> Accuracy: {accuracy_xgb:.4f}, AUC: {auc_xgb:.4f}\n")

    def test_regression(self):
        """Encapsulated test for the regression task with comparison."""
        print("\n" + "=" * 50 + "\n--- Running Regression Test ---\n" + "=" * 50)

        # Fix: make_regression may return 3 values, we only need X and y
        result = make_regression(
            n_samples=50000,
            n_features=20,
            n_informative=10,
            noise=25.0,
            random_state=42,
        )
        X_np, y_np = result[0], result[1]
        X, y = jnp.array(X_np, dtype=jnp.float32), jnp.array(y_np, dtype=jnp.float32)

        params = {
            "n_estimators": 10,
            "learning_rate": 0.1,
            "max_depth": 3,
            "max_bin": 64,
            "reg_lambda": 1.0,
            "gamma": 0.0,
            "min_child_weight": 1.0,
        }

        print("--- 1. Testing JAX Regressor ---")
        model_jax = XGBoostJAX(**params, objective="reg:squarederror")

        start_run = time.time()
        model_jax.fit(X, y)
        print(
            f"JAX Regressor compiled + first fit in: {time.time() - start_run:.4f} seconds"
        )
        start_run = time.time()
        model_jax.fit(X, y)
        print(f"JAX Regressor execution time: {time.time() - start_run:.4f} seconds")

        y_pred_jax = model_jax.predict(X)
        mse_jax = mean_squared_error(y, y_pred_jax)
        print(f"JAX Regressor MSE: {mse_jax:.4f}\n")

        print("--- 2. Testing Official XGBoost Regressor for comparison ---")
        xgb_params = params.copy()
        model_xgb = xgb.XGBRegressor(
            **xgb_params,
            tree_method="hist",
            objective="reg:squarederror",
            eval_metric="rmse",
        )

        start_run = time.time()
        model_xgb.fit(X_np, y_np)
        print(f"Official XGBoost fit time: {time.time() - start_run:.4f} seconds")
        y_pred_xgb = model_xgb.predict(X_np)
        mse_xgb = mean_squared_error(y_np, y_pred_xgb)
        print(f"Official XGBoost MSE: {mse_xgb:.4f}\n")

    def test_debug_functions(self):
        """Tests the debug functions for tree visualization and leaf prediction."""
        print("\n" + "=" * 50 + "\n--- Testing Debug Functions ---\n" + "=" * 50)

        # Import the debug functions
        from hist_jax import (
            count_samples_per_node,
            count_samples_per_node_ensemble,
            predict_ensemble_leaves,
            predict_tree_leaves,
            pretty_print_ensemble,
            print_leaf_predictions,
            print_node_sample_counts,
        )

        # Create a small dataset for testing
        X_np, y_np = make_classification(
            n_samples=21,
            n_features=4,
            n_informative=3,
            n_redundant=0,
            n_classes=2,
            random_state=42,
        )
        X = jnp.array(X_np, dtype=jnp.float32)
        y = jnp.array(y_np, dtype=jnp.float32)

        # Train a simple model
        params = {
            "n_estimators": 2,
            "learning_rate": 0.1,
            "max_depth": 2,
            "max_bin": 4,
            "reg_lambda": 1.0,
            "gamma": 0.0,
            "min_child_weight": 1.0,
        }

        model = XGBoostJAX(**params, objective="binary:logistic")
        model.fit(X, y)

        # Ensure model is fitted
        assert model.model is not None, "Model should be fitted"

        print("--- 1. Testing pretty_print_ensemble ---")
        pretty_print_ensemble(model.model)

        print("\n--- 2. Testing predict_tree_leaves ---")
        tree = model.model.trees[0]
        leaf_nodes = predict_tree_leaves(X[:5], tree)  # Test with first 5 samples
        print(f"Leaf nodes for first 5 samples in tree 0: {leaf_nodes}")

        # Verify the shape
        self.assertEqual(leaf_nodes.shape, (5,))
        # Verify all leaf nodes are valid indices
        max_node_idx = len(tree.feature) - 1
        self.assertTrue(jnp.all(leaf_nodes >= 0))
        self.assertTrue(jnp.all(leaf_nodes <= max_node_idx))

        print("\n--- 3. Testing predict_ensemble_leaves ---")
        ensemble_leaves = predict_ensemble_leaves(X[:5], model.model)
        print(f"Ensemble leaf shape: {ensemble_leaves.shape}")

        # Verify the shape: (n_trees, n_samples)
        self.assertEqual(ensemble_leaves.shape, (2, 5))

        print("\n--- 4. Testing print_leaf_predictions ---")
        print_leaf_predictions(
            X,
            model.model,
            y,
            max_samples=5,
            learning_rate=params["learning_rate"],
            objective="binary:logistic",
        )

        print("\n--- 5. Testing count_samples_per_node ---")
        tree_counts = count_samples_per_node(X, tree)
        print(f"Sample counts for tree 0: {tree_counts}")

        # Verify the shape: should match number of nodes in tree
        n_nodes = len(tree.feature)
        self.assertEqual(tree_counts.shape, (n_nodes,))

        # Verify total count equals number of samples (all samples should reach some leaf)
        total_count = jnp.sum(tree_counts)
        self.assertEqual(total_count, X.shape[0])

        # Verify only leaf nodes have non-zero counts
        for node_idx in range(n_nodes):
            if tree.is_leaf[node_idx]:
                # Leaf nodes can have zero or more samples
                self.assertTrue(tree_counts[node_idx] >= 0)
            else:
                # Split nodes should have zero samples (samples only stay at leaves)
                self.assertEqual(tree_counts[node_idx], 0)

        print("\n--- 6. Testing count_samples_per_node_ensemble ---")
        ensemble_counts = count_samples_per_node_ensemble(X, model.model)
        print(f"Ensemble counts shape: {ensemble_counts.shape}")

        # Verify the shape: (n_trees, n_nodes)
        n_trees = len(model.model.trees)
        self.assertEqual(ensemble_counts.shape, (n_trees, n_nodes))

        # Verify each tree's total count equals number of samples
        for tree_idx in range(n_trees):
            tree_total = jnp.sum(ensemble_counts[tree_idx])
            self.assertEqual(tree_total, X.shape[0])

        # Verify consistency with single tree function
        tree_0_counts_individual = count_samples_per_node(X, model.model.trees[0])
        tree_0_counts_ensemble = ensemble_counts[0]
        self.assertTrue(jnp.allclose(tree_0_counts_individual, tree_0_counts_ensemble))

        print("\n--- 7. Testing print_node_sample_counts ---")
        # Test for specific tree
        print("\n7a. Printing counts for Tree 0 (leaves only):")
        print_node_sample_counts(X, model.model, tree_idx=0, show_only_leaves=True)

        print("\n7b. Printing counts for Tree 0 (all nodes):")
        print_node_sample_counts(X, model.model, tree_idx=0, show_only_leaves=False)

        print("\n7c. Printing counts for all trees (leaves only):")
        print_node_sample_counts(X, model.model, tree_idx=None, show_only_leaves=True)

        # Test error handling for invalid tree index
        print("\n7d. Testing invalid tree index:")
        print_node_sample_counts(X, model.model, tree_idx=99, show_only_leaves=True)

        print("\n--- 8. Advanced validation ---")
        # Cross-validation: ensure consistency between different methods

        # Method 1: Using predict_tree_leaves + manual counting
        leaves_method1 = predict_tree_leaves(X, tree)
        unique_leaves, counts_method1 = jnp.unique(leaves_method1, return_counts=True)

        # Method 2: Using count_samples_per_node
        counts_method2 = count_samples_per_node(X, tree)

        # Verify that non-zero counts match
        for i, leaf_idx in enumerate(unique_leaves):
            expected_count = counts_method1[i]
            actual_count = counts_method2[leaf_idx]
            self.assertEqual(
                expected_count,
                actual_count,
                f"Mismatch for leaf {leaf_idx}: expected {expected_count}, got {actual_count}",
            )

        print("✅ All sample counting validations passed!")

        print("\n--- Debug Functions Test Completed Successfully ---")


if __name__ == "__main__":
    unittest.main()



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for expression printer module.
"""

import pytest

from mplang.core.dtype import FLOAT32, UINT64
from mplang.core.expr import (
    AccessExpr,
    CallExpr,
    CondExpr,
    ConvExpr,
    EvalExpr,
    FuncDefExpr,
    ShflExpr,
    ShflSExpr,
    TupleExpr,
    VariableExpr,
    WhileExpr,
)
from mplang.core.expr.printer import Printer
from mplang.core.mask import Mask
from mplang.core.mptype import MPType, Rank
from mplang.core.pfunc import PFunction
from mplang.core.tensor import TensorType


class TestPrinter:
    """Test the Printer class for expression visualization."""

    def test_printer_initialization(self):
        """Test basic printer initialization."""
        printer = Printer()
        assert printer.indent_size == 2
        assert printer._cur_indent == 0
        assert printer._output == []
        assert printer._visited == {}
        assert printer._counter == 0

    def test_printer_custom_indent(self):
        """Test printer with custom indent size."""
        printer = Printer(indent_size=4)
        assert printer.indent_size == 4

    def test_write_functionality(self):
        """Test write functionality with indentation."""
        printer = Printer(indent_size=2)

        # Test writing without indentation
        printer._write("test line")
        assert printer._output == ["test line"]

        # Test writing with indentation
        printer._cur_indent = 1
        printer._write("indented line")
        assert printer._output == ["test line", "  indented line"]


class TestPrinterExpressions:
    """Test printer for different expression types."""

    def test_variable_expr_printing(self, pmask_2p):
        """Test printing of VariableExpr."""
        printer = Printer(compact_format=False)
        mptype = MPType.tensor(FLOAT32, (2, 2), pmask_2p)
        expr = VariableExpr("test_var", mptype)

        result = printer.print_expr(expr)

        # Expected output format with exact literal comparison
        expected = '%0 = pname("test_var") : f32[2, 2]<3>'

        assert result == expected

    def test_eval_expr_printing(self, pmask_2p, pfunc_1i1o):
        """Test printing of EvalExpr."""
        printer = Printer()
        mptype = MPType.tensor(FLOAT32, (2, 3), pmask_2p)
        arg1 = VariableExpr("input", mptype)
        expr = EvalExpr(pfunc_1i1o, [arg1])

        result = printer.print_expr(expr)
        lines = result.split("\n")

        # Should contain the eval expression (in compact mode, no pname for variables)
        assert any("peval(" in line for line in lines)
        # Should contain fn_name since verbose=false still shows it
        assert any("fn_name=" in line for line in lines)

    def test_eval_expr_with_rmask_printing(self, pmask_2p, pfunc_1i1o):
        """Test printing of EvalExpr with rmask."""
        printer = Printer()
        mptype = MPType.tensor(FLOAT32, (2, 3), pmask_2p)
        arg = VariableExpr("input", mptype)
        rmask = Mask(3)  # 0b11
        expr = EvalExpr(pfunc_1i1o, [arg], rmask=rmask)

        result = printer.print_expr(expr)
        assert "rmask=" in result

    def test_eval_expr_verbose_peval_printing(self, pmask_2p):
        """Test printing of EvalExpr with verbose_peval enabled."""

        # Create a PFunction with actual fn_text and fn_type
        pfunc_with_text = PFunction(
            fn_type="jax",
            ins_info=[TensorType(FLOAT32, (2, 3))],
            outs_info=[TensorType(FLOAT32, (2, 3))],
            fn_name="add_one",
            fn_text="lambda x: x + 1",
        )

        # Test without verbose_peval (default)
        printer_normal = Printer()
        mptype = MPType.tensor(FLOAT32, (2, 3), pmask_2p)
        arg = VariableExpr("input", mptype)
        expr = EvalExpr(pfunc_with_text, [arg])

        result_normal = printer_normal.print_expr(expr)
        assert "fn_type=jax" in result_normal
        assert "fn_text=" not in result_normal
        assert "fn_name=add_one" in result_normal

        # Test with verbose_peval enabled
        printer_verbose = Printer(verbose_peval=True)
        result_verbose = printer_verbose.print_expr(expr)
        assert "fn_name=add_one" in result_verbose
        assert "fn_text=lambda x: x + 1" in result_verbose
        assert "fn_type=jax" in result_verbose

    def test_tuple_expr_printing(self, pmask_2p):
        """Test printing of TupleExpr."""
        printer = Printer()
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        arg1 = VariableExpr("x", mptype)
        arg2 = VariableExpr("y", mptype)
        expr = TupleExpr([arg1, arg2])

        result = printer.print_expr(expr)
        lines = result.split("\n")

        # Should contain tuple expression (in compact mode, uses variable names)
        assert any("tuple(x, y)" in line for line in lines)

    def test_conv_expr_printing(self):
        """Test printing of ConvExpr."""
        printer = Printer()
        # Use disjoint pmasks: party 0 and party 1
        mptype1 = MPType.tensor(UINT64, (), Mask(1))  # 0b01 - party 0
        mptype2 = MPType.tensor(UINT64, (), Mask(2))  # 0b10 - party 1
        var1 = VariableExpr("p0_var", mptype1)
        var2 = VariableExpr("p1_var", mptype2)
        expr = ConvExpr([var1, var2])

        result = printer.print_expr(expr)
        lines = result.split("\n")

        # Should contain conv expression (in compact mode, uses variable names)
        assert any("pconv(p0_var, p1_var)" in line for line in lines)

    def test_multiple_expressions_same_name(self, pmask_2p):
        """Test that multiple expressions of the same type get different names."""
        printer = Printer(compact_format=False)
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        # Two different expression instances with the same name
        expr1 = VariableExpr("x", mptype)
        expr2 = VariableExpr("x", mptype)

        # Create a tuple expression with both
        tuple_expr = TupleExpr([expr1, expr2])

        result = printer.print_expr(tuple_expr)

        # Expected output format with exact literal comparison
        # The printer should assign different SSA names (%0, %1) to the two different expressions.
        expected = """
%0 = pname("x") : u64<3>
%1 = pname("x") : u64<3>
%2 = tuple(%0, %1) : (u64<3>, u64<3>)
""".strip()

        assert result == expected

    def test_shfl_s_expr_printing(self, pmask_2p):
        """Test printing of ShflSExpr."""
        printer = Printer()
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        src_val = VariableExpr("src", mptype)
        # Use a larger target pmask that requires 3 src_ranks
        target_pmask = Mask(7)  # 0b111 for 3 parties
        # Need 3 src_ranks since target_pmask has 3 bits set
        # Source has pmask_2p (parties 0,1), so src_ranks must be from those parties
        src_ranks = [Rank(0), Rank(1), Rank(0)]  # Party 2 gets data from party 0
        expr = ShflSExpr(src_val, target_pmask, src_ranks)

        result = printer.print_expr(expr)
        lines = result.split("\n")

        # Should contain pshfl_s expression
        assert any("pshfl_s(" in line for line in lines)
        assert any("pmask=" in line for line in lines)
        assert any("src_ranks=" in line for line in lines)

    def test_shfl_expr_printing(self, pmask_2p):
        """Test printing of ShflExpr."""
        printer = Printer()
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        src = VariableExpr("src", mptype)
        index = VariableExpr("idx", mptype)
        expr = ShflExpr(src, index)

        result = printer.print_expr(expr)
        lines = result.split("\n")

        # Should contain pshfl expression
        assert any("pshfl(" in line for line in lines)

    def test_access_expr_printing(self, pmask_2p):
        """Test printing of AccessExpr."""
        printer = Printer(compact_format=False)
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        base_expr = VariableExpr("base", mptype)
        expr = AccessExpr(base_expr, 0)

        result = printer.print_expr(expr)
        lines = result.split("\n")

        # Should contain access expression with index attribute
        assert any("access(" in line for line in lines)
        assert any("index=0" in line for line in lines)

    def test_func_def_expr_printing(self, pmask_2p):
        """Test printing of FuncDefExpr with parameter usage."""
        printer = Printer(compact_format=False)

        # Create a function body that actually uses the parameters
        x_mptype = MPType.tensor(FLOAT32, (1,), pmask_2p)
        y_mptype = MPType.tensor(FLOAT32, (1,), pmask_2p)

        x_var = VariableExpr("x", x_mptype)
        y_var = VariableExpr("y", y_mptype)

        # Tuple the two parameters together
        body = TupleExpr([x_var, y_var])
        func_def = FuncDefExpr(["x", "y"], body)

        result = printer.print_expr(func_def)

        # Expected output format with exact literal comparison
        expected = """
(x, y) {
  %0 = pname("x") : f32[1]<3>
  %1 = pname("y") : f32[1]<3>
  %2 = tuple(%0, %1) : (f32[1]<3>, f32[1]<3>)
  return %2
}
"""

        assert result == expected.strip()


class TestPrinterComplexExpressions:
    """Test printer for complex nested expressions."""

    def test_cond_expr_printing(self, pmask_2p):
        """Test printing of CondExpr with nested functions that actually use their parameters."""
        printer = Printer(compact_format=False)

        # Create predicate
        pred_mptype = MPType.tensor(UINT64, (), pmask_2p)
        pred = VariableExpr("pred", pred_mptype)

        # Create then function that actually uses the parameter 'x'
        x_mptype = MPType.tensor(FLOAT32, (1,), pmask_2p)
        x_var = VariableExpr("x", x_mptype)  # Reference the parameter
        then_fn = FuncDefExpr(["x"], x_var)

        # Create else function that also uses the parameter 'x'
        else_fn = FuncDefExpr(["x"], x_var)

        # Create arguments
        arg_mptype = MPType.tensor(FLOAT32, (1,), pmask_2p)
        arg = VariableExpr("arg", arg_mptype)

        # Create conditional expression
        expr = CondExpr(pred, then_fn, else_fn, [arg])

        result = printer.print_expr(expr)

        # Expected output format with exact literal comparison
        expected = """
%0 = pname("pred") : u64<3>
%1 = pname("arg") : f32[1]<3>
%2 = pcond(%0, %1) {
  then_fn: (x) {
    %0 = pname("x") : f32[1]<3>
    return %0
  }
  else_fn: (x) {
    %0 = pname("x") : f32[1]<3>
    return %0
  }
} : f32[1]<3>
"""

        assert result == expected.strip()

    def test_all_expr_types_printing(self, pmask_2p, pfunc_1i1o):
        """Test printing of a complex expression involving all Expr types with meaningful parameter usage."""
        printer = Printer(compact_format=False, inline_pcall=False)

        # 1. Variable expressions
        var1_mptype = MPType.tensor(FLOAT32, (2,), pmask_2p)
        var1 = VariableExpr("var1", var1_mptype)

        var2_mptype = MPType.tensor(UINT64, (2,), pmask_2p)
        var2 = VariableExpr("var2", var2_mptype)

        var3_mptype = MPType.tensor(UINT64, (), pmask_2p)
        var3 = VariableExpr("var3", var3_mptype)

        # 2. Access expression
        access_expr = AccessExpr(var1, 0)

        # 3. Conv expression (requires disjoint masks)
        var_p0 = VariableExpr("p0", MPType.tensor(UINT64, (), Mask(1)))
        var_p1 = VariableExpr("p1", MPType.tensor(UINT64, (), Mask(2)))
        conv_expr = ConvExpr([var_p0, var_p1])

        # 4. Eval expression
        eval_expr = EvalExpr(pfunc_1i1o, [var1])

        # 5. Shuffle expressions
        shfl_expr = ShflExpr(var3, var3)
        shfl_s_expr = ShflSExpr(
            var3,
            Mask(7),  # target 3 parties
            [Rank(0), Rank(1), Rank(0)],  # src ranks for each target party
        )

        # 6. Function definition and call
        param_type = MPType.tensor(FLOAT32, (2,), pmask_2p)
        var_expr = VariableExpr("input_data", param_type)
        func_body = TupleExpr([var_expr, access_expr])
        func_def = FuncDefExpr(["input_data"], func_body)
        call_expr = CallExpr("test", func_def, [var1])

        # Access the first output of the call expression to get a single-output expr
        call_expr_first = AccessExpr(call_expr, 0)

        # 7. Build final comprehensive expression
        final_expr = TupleExpr([
            var1,  # VariableExpr
            var2,  # VariableExpr
            var3,  # VariableExpr
            access_expr,  # AccessExpr
            conv_expr,  # ConvExpr
            eval_expr,  # EvalExpr
            shfl_expr,  # ShflExpr
            shfl_s_expr,  # ShflSExpr
            call_expr_first,  # AccessExpr of CallExpr (includes FuncDefExpr, VariableExpr, TupleExpr)
        ])

        result = printer.print_expr(final_expr)

        # Expected output with exact format (SSA semantics - no duplicate definitions)
        expected = """
%0 = pname("var1") : f32[2]<3>
%1 = pname("var2") : u64[2]<3>
%2 = pname("var3") : u64<3>
%3 = access(%0) {index=0} : f32[2]<3>
%4 = pname("p0") : u64<1>
%5 = pname("p1") : u64<2>
%6 = pconv(%4, %5) : u64<3>
%7 = peval(%0) {fn_type=mock, fn_name=mock_unary} : f32[2, 3]<3>
%8 = pshfl(%2, %2) : u64
%9 = pshfl_s(%2) {pmask=7, src_ranks=[0, 1, 0]} : u64<7>
%10 = pcall(%0) {
  fn: (input_data) {
    %0 = pname("input_data") : f32[2]<3>
    %1 = pname("var1") : f32[2]<3>
    %2 = access(%1) {index=0} : f32[2]<3>
    %3 = tuple(%0, %2) : (f32[2]<3>, f32[2]<3>)
    return %3
  }
} : (f32[2]<3>, f32[2]<3>)
%11 = access(%10) {index=0} : f32[2]<3>
%12 = tuple(%0, %1, %2, %3, %6, %7, %8, %9, %11) : (f32[2]<3>, u64[2]<3>, u64<3>, f32[2]<3>, u64<3>, f32[2, 3]<3>, u64, u64<7>, f32[2]<3>)
"""

        assert result == expected.strip()

        # Test with optimize_variables=True for comparison
        printer_optimized = Printer(compact_format=True, inline_pcall=False)
        result_optimized = printer_optimized.print_expr(final_expr)

        # Expected output with variable optimization (compact mode uses variable names directly)
        expected_optimized = """
%0 = pconv(p0, p1) : u64<3>
%1 = peval(var1) {fn_type=mock, fn_name=mock_unary} : f32[2, 3]<3>
%2 = pshfl(var3, var3) : u64
%3 = pshfl_s(var3) {pmask=7, src_ranks=[0, 1, 0]} : u64<7>
%4 = pcall(var1) {
  fn: (input_data) {
    %0 = tuple(input_data, var1) : (f32[2]<3>, f32[2]<3>)
    return %0
  }
} : (f32[2]<3>, f32[2]<3>)
%5 = tuple(var1, var2, var3, var1, %0, %1, %2, %3, %4:0) : (f32[2]<3>, u64[2]<3>, u64<3>, f32[2]<3>, u64<3>, f32[2, 3]<3>, u64, u64<7>, f32[2]<3>)
"""

        assert result_optimized == expected_optimized.strip()

    def test_while_expr_printing(self, pmask_2p):
        """Test printing of WhileExpr with nested functions that actually use their parameters."""
        printer = Printer(compact_format=False)

        # Create condition function that uses the 'state' parameter
        state_mptype = MPType.tensor(FLOAT32, (1,), pmask_2p)
        state_var = VariableExpr("state", state_mptype)
        # Access the first element of state to check if we should continue
        cond_body = AccessExpr(state_var, 0)
        cond_fn = FuncDefExpr(["state"], cond_body)

        # Create body function that also uses the 'state' parameter
        # Return state itself (identity transformation)
        body_fn = FuncDefExpr(["state"], state_var)

        # Create initial value
        init_mptype = MPType.tensor(UINT64, (), pmask_2p)
        init = VariableExpr("init", init_mptype)

        # Create while expression
        expr = WhileExpr(cond_fn, body_fn, [init])

        result = printer.print_expr(expr)

        # Expected output format with exact literal comparison
        expected = """
%0 = pname("init") : u64<3>
%1 = pwhile(%0) {
  cond_fn: (state) {
    %0 = pname("state") : f32[1]<3>
    %1 = access(%0) {index=0} : f32[1]<3>
    return %1
  }
  body_fn: (state) {
    %0 = pname("state") : f32[1]<3>
    return %0
  }
} : f32[1]<3>
"""

        assert result == expected.strip()

    def test_call_expr_printing(self, pmask_2p):
        """Test printing of CallExpr with nested function that uses its parameters."""
        printer = Printer(compact_format=False, inline_pcall=False)

        # Create function body that actually uses the parameters
        x_mptype = MPType.tensor(FLOAT32, (1,), pmask_2p)
        y_mptype = MPType.tensor(FLOAT32, (1,), pmask_2p)

        x_var = VariableExpr("x", x_mptype)
        y_var = VariableExpr("y", y_mptype)

        # Create a function that tuples its arguments together
        fn_body = TupleExpr([x_var, y_var])
        fn = FuncDefExpr(["x", "y"], fn_body)

        # Create arguments
        arg1_mptype = MPType.tensor(UINT64, (), pmask_2p)
        arg2_mptype = MPType.tensor(UINT64, (), pmask_2p)
        arg1 = VariableExpr("arg1", arg1_mptype)
        arg2 = VariableExpr("arg2", arg2_mptype)

        # Create call expression
        expr = CallExpr("test", fn, [arg1, arg2])

        result = printer.print_expr(expr)

        # Expected output format with exact literal comparison
        expected = """
%0 = pname("arg1") : u64<3>
%1 = pname("arg2") : u64<3>
%2 = pcall(%0, %1) {
  fn: (x, y) {
    %0 = pname("x") : f32[1]<3>
    %1 = pname("y") : f32[1]<3>
    %2 = tuple(%0, %1) : (f32[1]<3>, f32[1]<3>)
    return %2
  }
} : (f32[1]<3>, f32[1]<3>)
"""

        assert result == expected.strip()

    def test_nested_expressions_printing(self, pmask_2p):
        """Test printing of deeply nested expressions."""
        printer = Printer(compact_format=False)

        # Create a complex nested structure
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        inner = VariableExpr("inner", mptype)
        access1 = AccessExpr(inner, 0)
        access2 = AccessExpr(access1, 0)

        result = printer.print_expr(access2)

        # Expected output format with exact literal comparison
        expected = """
%0 = pname("inner") : u64<3>
%1 = access(%0) {index=0} : u64<3>
%2 = access(%1) {index=0} : u64<3>
"""

        assert result == expected.strip()


class TestPrinterIndentation:
    """Test printer indentation behavior."""

    def test_indentation_reset_between_prints(self, pmask_2p):
        """Test that indentation is reset between print calls."""
        printer = Printer(compact_format=False)  # Use non-compact to see pname output

        # Create a complex expression that modifies indentation
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        body = VariableExpr("x", mptype)
        func_def = FuncDefExpr(["x"], body)

        # Print first expression
        printer.print_expr(func_def)

        # Print second expression - indentation should be reset
        simple_expr = VariableExpr("y", mptype)
        result2 = printer.print_expr(simple_expr)

        # Second result should not have indentation from first
        assert result2.strip() == '%0 = pname("y") : u64<3>'

    def test_counter_reset_between_prints(self, pmask_2p):
        """Test that expression counter is reset between print calls."""
        printer = Printer(compact_format=False)  # Use non-compact to see pname output

        # Print first expression
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        expr1 = VariableExpr("x", mptype)
        result1 = printer.print_expr(expr1)
        assert '%0 = pname("x")' in result1

        # Print second expression - counter should reset
        expr2 = VariableExpr("y", mptype)
        result2 = printer.print_expr(expr2)
        assert '%0 = pname("y")' in result2

    def test_expression_names_reset_between_prints(self, pmask_2p):
        """Test that expression names are reset between print calls."""
        printer = Printer(compact_format=False)  # Use non-compact to see pname output

        # Print first expression
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        expr1 = VariableExpr("x", mptype)
        result1 = printer.print_expr(expr1)
        # After print_expr, _visited is reset so we can't check it
        # But we can verify the output shows %0
        assert '%0 = pname("x")' in result1

        # Print second expression - names should reset
        expr2 = VariableExpr("y", mptype)
        result2 = printer.print_expr(expr2)
        # Both should start with %0 because counter resets
        assert '%0 = pname("y")' in result2


class TestPrinterEdgeCases:
    """Test printer edge cases and error conditions."""

    def test_empty_tuple_expr(self):
        """Test printing of TupleExpr with no arguments."""
        printer = Printer()
        expr = TupleExpr([])

        result = printer.print_expr(expr)
        assert "%0 = tuple()" in result

    def test_printer_output_format(self, pmask_2p):
        """Test that printer output format is consistent."""
        printer = Printer(compact_format=False)  # Use non-compact to see pname output

        # Test multiple expressions to verify format consistency
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        expr1 = VariableExpr("x", mptype)
        result1 = printer.print_expr(expr1)

        # Should start with %0 and be properly formatted
        assert result1.strip() == '%0 = pname("x") : u64<3>'

        # Test that each new print starts fresh
        expr2 = VariableExpr("y", mptype)
        result2 = printer.print_expr(expr2)
        assert result2.strip() == '%0 = pname("y") : u64<3>'

    def test_deep_nesting_indentation(self, pmask_2p):
        """Test printer with deeply nested expressions that use parameters meaningfully."""
        printer = Printer(compact_format=False, inline_pcall=False)

        # Create nested function definitions with meaningful parameter usage
        inner_param_type = MPType.tensor(FLOAT32, (1,), pmask_2p)
        middle_param_type = MPType.tensor(FLOAT32, (1,), pmask_2p)

        # Inner function: takes parameter and accesses its first element
        inner_param = VariableExpr("inner_param", inner_param_type)
        inner_body = AccessExpr(inner_param, 0)
        inner_fn = FuncDefExpr(["inner_param"], inner_body)

        # Middle function: takes parameter and calls inner function with it
        middle_param = VariableExpr("middle_param", middle_param_type)
        middle_body = CallExpr("middle", inner_fn, [middle_param])
        middle_fn = FuncDefExpr(["middle_param"], middle_body)

        # Outer expression: call middle function with a variable expression
        arg_mptype = MPType.tensor(UINT64, (), pmask_2p)
        arg = VariableExpr("arg", arg_mptype)
        outer_expr = CallExpr("outer", middle_fn, [arg])

        result = printer.print_expr(outer_expr)

        # Expected output format with exact literal comparison
        # This shows the deep nesting with proper indentation and meaningful parameter usage
        expected = """
%0 = pname("arg") : u64<3>
%1 = pcall(%0) {
  fn: (middle_param) {
    %0 = pname("middle_param") : f32[1]<3>
    %1 = pcall(%0) {
      fn: (inner_param) {
        %0 = pname("inner_param") : f32[1]<3>
        %1 = access(%0) {index=0} : f32[1]<3>
        return %1
      }
    } : f32[1]<3>
    return %1
  }
} : f32[1]<3>
"""

        assert result == expected.strip()

    def test_function_def_params_handling(self, pmask_2p):
        """Test function definition parameter handling."""
        printer = Printer(compact_format=False)

        # Test with multiple parameters
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        body = VariableExpr("param1", mptype)
        func_def = FuncDefExpr(["param1", "param2", "param3"], body)

        result = printer.print_expr(func_def)
        lines = result.split("\n")

        # Should contain all parameters
        func_line = next(line for line in lines if "(" in line and "param1" in line)
        assert "param1, param2, param3" in func_line

    def test_access_expr_with_different_indices(self, pmask_2p):
        """Test AccessExpr with various index values."""
        printer = Printer(compact_format=False)

        mptype = MPType.tensor(UINT64, (), pmask_2p)
        base_expr = VariableExpr("base", mptype)

        # Test index 0 (the only valid index for single output expressions)
        access_expr = AccessExpr(base_expr, 0)
        result = printer.print_expr(access_expr)
        assert "index=0" in result
        assert "access(" in result

    def test_printer_with_single_expression_multiple_outputs(self, pmask_2p):
        """Test printer behavior with expressions that might have multiple outputs."""
        printer = Printer(compact_format=False)

        # Create a complex expression tree
        mptype = MPType.tensor(UINT64, (), pmask_2p)
        expr1 = VariableExpr("x", mptype)
        expr2 = VariableExpr("y", mptype)
        tuple_expr = TupleExpr([expr1, expr2])
        access_expr = AccessExpr(tuple_expr, 0)

        result = printer.print_expr(access_expr)
        lines = result.split("\n")

        # Should show the dependency chain clearly
        assert len(lines) >= 3
        assert any("pname(" in line for line in lines)
        assert any("tuple(%0, %1)" in line for line in lines)
        assert any("access(" in line and "index=0" in line for line in lines)


class TestPrinterStringMethods:
    """Test printer internal string handling methods."""

    def test_write_with_multiline_content(self):
        """Test _write method behavior."""
        printer = Printer()

        # Test single line
        printer._write("single line")
        assert printer._output == ["single line"]

        # Test with indentation
        printer._cur_indent = 1
        printer._write("indented line")
        assert printer._output == ["single line", "  indented line"]

        # Test multiple writes
        printer._write("another line")
        assert len(printer._output) == 3
        assert printer._output[-1] == "  another line"


class TestPrinterMeaningfulParameterUsage:
    """Test printer with meaningful parameter usage scenarios."""

    def test_meaningful_parameter_usage_scenarios(self, pmask_2p):
        """Test various scenarios where function parameters are meaningfully used."""
        printer = Printer(compact_format=False)

        # Scenario 1: Parameter used in conditional predicate
        param_type = MPType.tensor(FLOAT32, (2,), pmask_2p)
        param_var = VariableExpr("data", param_type)

        # Use parameter in both branches - return the same parameter
        then_body = param_var  # Return parameter directly
        else_body = param_var  # Return parameter directly

        pred_mptype = MPType.tensor(UINT64, (), pmask_2p)
        pred = VariableExpr("pred", pred_mptype)
        then_fn = FuncDefExpr(["data"], then_body)
        else_fn = FuncDefExpr(["data"], else_body)

        arg_mptype = MPType.tensor(FLOAT32, (2,), pmask_2p)
        arg = VariableExpr("arg", arg_mptype)
        cond_expr = CondExpr(pred, then_fn, else_fn, [arg])

        result = printer.print_expr(cond_expr)

        expected = """
%0 = pname("pred") : u64<3>
%1 = pname("arg") : f32[2]<3>
%2 = pcond(%0, %1) {
  then_fn: (data) {
    %0 = pname("data") : f32[2]<3>
    return %0
  }
  else_fn: (data) {
    %0 = pname("data") : f32[2]<3>
    return %0
  }
} : f32[2]<3>
"""

        assert result == expected.strip()

    def test_parameter_reuse_in_complex_expressions(self, pmask_2p):
        """Test parameter being used multiple times in the same function."""
        printer = Printer(compact_format=False)

        # Create a function that uses the same parameter multiple times
        param_type = MPType.tensor(FLOAT32, (1,), pmask_2p)
        x_var = VariableExpr("x", param_type)

        # Use x multiple times: tuple x with itself
        func_body = TupleExpr([x_var, x_var])
        func_def = FuncDefExpr(["x"], func_body)

        result = printer.print_expr(func_def)

        # The same variable expression is now properly cached (SSA semantics)
        expected = """
(x) {
  %0 = pname("x") : f32[1]<3>
  %1 = tuple(%0, %0) : (f32[1]<3>, f32[1]<3>)
  return %1
}
"""

        assert result == expected.strip()

    def test_while_loop_with_meaningful_state_usage(self, pmask_2p):
        """Test while loop where state parameter is meaningfully used."""
        printer = Printer(compact_format=False)

        # Create a while loop that actually processes the state
        state_type = MPType.tensor(FLOAT32, (3,), pmask_2p)
        state_var = VariableExpr("state", state_type)

        # Condition: check if state[0] is non-zero (simplified)
        cond_body = AccessExpr(state_var, 0)
        cond_fn = FuncDefExpr(["state"], cond_body)

        # Body: return the same state (using index 0)
        body_body = AccessExpr(state_var, 0)
        body_fn = FuncDefExpr(["state"], body_body)

        # Initial state
        init_mptype = MPType.tensor(UINT64, (), pmask_2p)
        init = VariableExpr("init", init_mptype)

        while_expr = WhileExpr(cond_fn, body_fn, [init])
        result = printer.print_expr(while_expr)

        expected = """
%0 = pname("init") : u64<3>
%1 = pwhile(%0) {
  cond_fn: (state) {
    %0 = pname("state") : f32[3]<3>
    %1 = access(%0) {index=0} : f32[3]<3>
    return %1
  }
  body_fn: (state) {
    %0 = pname("state") : f32[3]<3>
    %1 = access(%0) {index=0} : f32[3]<3>
    return %1
  }
} : f32[3]<3>
"""

        assert result == expected.strip()


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import ibis
import numpy.testing as npt
import pandas as pd
import pytest

# you can switch engine in ["duckdb", "substrait"]
# some testing may crash in "substrait"
_ibis_engine = "duckdb"


class TestIbis:
    def test_scalar_udf(self):
        def uppercase_string(s: str) -> str:
            return s.upper()

        ibis_uppercase_string = ibis.udf.scalar.builtin(name="uppercase_string")(
            uppercase_string
        )

        t = ibis.table(schema={"col": "str"}, name="table")
        expr = t.select(upper_str=ibis_uppercase_string(t["col"]))

        data = {"col": ["hello", "world"]}
        res = self._eval_duckdb(expr, [data], udfs=[uppercase_string])
        print(res)

    @pytest.mark.parametrize(
        "train_size,test_size,shuffle,random_state",
        [
            (0.75, None, False, None),
            (10, None, False, None),
            # (None, 0.35, True, None),
            # (None, 30, True, None),
        ],
    )
    def test_train_test_split(self, train_size, test_size, shuffle, random_state):
        if random_state is None:
            random_state = 1024
        if train_size is None and test_size is None:
            raise ValueError("train_size and test_size cannot all be None")

        data = {
            "id": [f"id{i}" for i in range(100)],
            "a1": [f"a{i}" for i in range(100)],
            "a2": list(range(100)),
        }

        schema = {"id": "str", "a1": "str", "a2": "int"}
        t = self._memtable(data, schema=schema)
        total = t.count().execute()

        if train_size is not None and train_size <= 1.0:
            train_size = int(total * train_size)
        if test_size is not None and test_size <= 0:
            test_size = int(total * test_size)
        if train_size is None:
            train_size = total - test_size
        if test_size is None:
            test_size = total - train_size
        if train_size + test_size > total:
            raise ValueError(f"split size overflow: {train_size}, {test_size}, {total}")

        # shuffle will crash.
        # duckdb.duckdb.InvalidInputException: Invalid Input Error:
        # More than one row returned by a subquery used as an expression - scalar subqueries can only return a single row.
        # if shuffle:
        #     # TODO: random_state
        #     t = t.order_by(ibis.random())

        train_data = t.limit(train_size).execute()
        test_data = t.limit(test_size, offset=train_size).execute()
        train_data = train_data.reset_index(drop=True)
        test_data = test_data.reset_index(drop=True)
        print(
            f"totol_size: {total}, train_size:{train_size}, test_size:{test_size}, shuffle: {shuffle}"
        )
        print(f"train_data: {train_data.shape} \n {train_data.head(5)}")
        print(f"test_data:  {test_data.shape}  \n {test_data.head(5)}")

    def test_join(self):
        schema1 = {"id1": "str", "id2": "str", "a1": "int", "a2": "bool"}
        schema2 = {"id3": "str", "id4": "str", "b1": "float"}

        keys1 = ["id1", "id2"]
        keys2 = ["id3", "id4"]

        def _join(t0: ibis.Table, t1: ibis.Table) -> ibis.Table:
            cond = ibis.and_(*[
                getattr(t0, k1) == getattr(t1, k2)
                for k1, k2 in zip(keys1, keys2, strict=True)
            ])
            return t0.join(t1, cond, how="inner")

        t0 = ibis.table(schema=schema1, name="table0")
        t1 = ibis.table(schema=schema2, name="table1")
        aligned = _join(t0, t1)
        data1 = {
            "id1": ["k1", "k2"],
            "id2": ["k3", "k4"],
            "a1": [1, 2],
            "a2": [True, False],
        }
        data2 = {
            "id3": ["k1", "k2", "ka"],
            "id4": ["k3", "k5", "kb"],
            "b1": [1.1, 2.1, 3.1],
        }
        res = self._eval(aligned, data1, data2, table_names=["table0", "table1"])
        print(f"join res: \n {res}")

    def test_union(self):
        schema = {"id": "str", "f0": "str", "f1": "int"}
        t1 = ibis.table(schema=schema, name="table1")
        t2 = ibis.table(schema=schema, name="table2")
        expr = t1.union(t2, distinct=False)
        data1 = {"id": ["ID1", "ID2"], "f0": ["x1", "x2"], "f1": [1, 2]}
        data2 = {"id": ["ID3", "ID4"], "f0": ["x3", "x4"], "f1": [3, 4]}
        df1 = pd.DataFrame(data1)
        df2 = pd.DataFrame(data2)

        res = self._eval(expr, data1, data2, table_names=["table1", "table2"])
        print(f"union res: \n{res}")
        combined_df = pd.concat([df1, df2], ignore_index=True)
        npt.assert_equal(res, combined_df.to_numpy())

    @pytest.mark.parametrize("op", ["+", "-", "*", "/"])
    @pytest.mark.parametrize(
        "features,new_feature_name",
        [
            (["a", "b"], "r"),
            (["a", "b"], "a"),
            (["a", "a"], "r"),
            (["a", "a"], "a"),
            (["a", "c"], "r"),
        ],
    )
    def test_binary_op(self, op: str, features, new_feature_name):
        t = ibis.table(schema={"a": "int", "b": "int", "c": "float"}, name="table")

        match op:
            case "+":
                result_expr = t[features[0]] + t[features[1]]
            case "-":
                result_expr = t[features[0]] - t[features[1]]
            case "*":
                result_expr = t[features[0]] * t[features[1]]
            case "/":
                result_expr = t[features[0]] / t[features[1]]
            case _:
                raise ValueError(f"Unsupported operation: {op}")

        new_table = t.mutate(**{new_feature_name: result_expr})

        assert new_feature_name in new_table.schema()
        res = self._eval(
            new_table, {"a": [1, 2, 3], "b": [4, 5, 6], "c": [4.0, 5.0, 6.0]}
        )
        print(f"{new_feature_name} = {features[0]} {op} {features[1]}:\n {res}")

    @pytest.mark.parametrize("columns", [["a", "b", "c", "d"]])
    @pytest.mark.parametrize("astype", ["int", "float", "string"])
    def test_cast(self, columns: list[str], astype: str):
        t = ibis.table(
            schema={"a": "int", "b": "float", "c": "string", "d": "bool"}, name="table"
        )
        # exprs = {col: t[col].cast(astype) for col in columns}
        exprs = {}
        for name in columns:
            col = t[name]
            # substrait donot support strip/trim and try_cast
            if col.type().is_string():
                col = col.strip()
            exprs[name] = col.cast(astype)
        new_t = t.mutate(**exprs)
        input = {
            "a": [1, 2, 3],
            "b": [4.1, 5.1, 6.1],
            "c": ["7", "8", "9"],  # "8.1" -> 8 will panic
            "d": [True, False, True],
        }
        res = self._eval(new_t, input)
        print(f"columns:{columns} cast to {astype}\n {res}")

    @pytest.mark.parametrize(
        "features,op,operands",
        [
            (["a1", "a2", "b1"], "standardize", []),
            (["a1", "a2", "b1"], "normalize", []),
            (["a1", "b1"], "range_limit", ["1", "2"]),
            (["a1", "b1"], "unary", ["+", "+", "1"]),  # unary_+
            (["a1", "b1"], "unary", ["+", "-", "1"]),  # unary_-
            (["a1", "b1"], "unary", ["-", "-", "1"]),  # unary_reverse_-
            (["a1", "b1"], "unary", ["+", "*", "2"]),  # unary_*
            (["a1", "b1"], "unary", ["+", "/", "3"]),  # unary_/
            (["a1", "b1"], "reciprocal", []),
            (["a1", "b1"], "round", []),
            (["a1", "b1"], "log_round", ["10"]),
            (["a1", "b1"], "log", ["e", "10"]),
            (["a1", "b1"], "log", ["2", "10"]),
            (["b1"], "sqrt", []),
            (["a1", "b1"], "exp", []),
            (["a3"], "length", []),
            (["a3"], "substr", ["0", "2"]),
        ],
    )
    def test_feature_calculate(self, features: list[str], op: str, operands: list[str]):
        t: ibis.Table = ibis.table(
            schema={"a1": "float", "a2": "float", "a3": "string", "b1": "int"},
            name="table",
        )

        def _standardize(col: ibis.Column):
            mean_expr = col.mean()
            std_expr = col.std()
            return ibis.cases((std_expr == 0, 0), else_=(col - mean_expr) / std_expr)

        def _normalize(col: ibis.Column):
            # norm = (x-min)/(ma-min)
            min_expr = col.min()
            max_expr = col.max()

            denominator = max_expr - min_expr

            return ibis.cases(
                (denominator == 0, 0), else_=(col - min_expr) / denominator
            )

        def _range_limit(col: ibis.Column):
            op_cnt = len(operands)
            if op_cnt != 2:
                raise ValueError(
                    f"range limit operator need 2 operands, but got {op_cnt}"
                )
            op0 = float(operands[0])
            op1 = float(operands[1])
            if op0 > op1:
                raise ValueError(
                    f"range limit operator expect min <= max, but get [{op0}, {op1}]"
                )
            new_col = ibis.cases((col < op0, op0), (col > op1, op1), else_=col)
            return new_col

        def _unary(col: ibis.Column):
            op_cnt = len(operands)
            if op_cnt != 3:
                raise ValueError(f"unary operator needs 3 operands, but got {op_cnt}")
            op0 = operands[0]
            if op0 not in ["+", "-"]:
                raise ValueError(f"unary op0 should be [+ -], but get {op0}")
            op1 = operands[1]
            if op1 not in ["+", "-", "*", "/"]:
                raise ValueError(f"unary op1 should be [+ - * /], but get {op1}")
            op3 = float(operands[2])
            if op1 == "+":
                new_col = col + op3
            elif op1 == "-":
                new_col = col - op3 if op0 == "+" else op3 - col
            elif op1 == "*":
                new_col = col * op3
            elif op1 == "/":
                if op0 == "+":
                    if op3 == 0:
                        raise ValueError("unary operator divide zero")
                    new_col = col / op3
                else:
                    new_col = op3 / col
            return new_col

        def _reciprocal(col: ibis.Column):
            return 1 / col

        def _round(col: ibis.Column):
            return col.round()

        def _log_round(col: ibis.Column):
            op_cnt = len(operands)
            if op_cnt != 1:
                raise ValueError(f"log operator needs 1 operands, but got {op_cnt}")
            op0 = float(operands[0])
            return (col + op0).log(2).round(2)

        def _log(col: ibis.Column):
            import math

            op_cnt = len(operands)
            if op_cnt != 2:
                raise ValueError(f"log operator needs 2 operands, but got {op_cnt}")

            op0 = operands[0]
            op1 = float(operands[1])
            adjusted_col = col + op1

            if op0 == "e":
                # 自然对数:log_e(x) = ln(x)
                new_col = adjusted_col.log(math.e)
            else:
                # 任意底数对数:log_base(x)
                base = float(op0)
                new_col = adjusted_col.log(base)

            return new_col

        def _sqrt(col: ibis.Column):
            return col.sqrt()

        def _exp(col: ibis.Column):
            return col.exp()

        def _length(col: ibis.Column):
            return col.length()

        def _substr(col: ibis.Column):
            op_cnt = len(operands)
            if op_cnt != 2:
                raise ValueError(f"substr operator needs 2 operands, but got {op_cnt}")
            start = int(operands[0])
            length = int(operands[1])
            return col.substr(start + 1, length)

        fn_map = {
            "standardize": _standardize,
            "normalize": _normalize,
            "range_limit": _range_limit,
            "unary": _unary,
            "reciprocal": _reciprocal,
            "round": _round,
            "log_round": _log_round,
            "log": _log,
            "sqrt": _sqrt,
            "exp": _exp,
            "length": _length,
            "substr": _substr,
        }
        fn = fn_map[op]
        exprs = {name: fn(t[name]) for name in features}
        new_t = t.mutate(**exprs)

        input = {
            "a1": [i * (-0.8) for i in range(3)],
            "a2": [0.1] * 3,
            "a3": ["AAA", "BBB", "CCC"],
            "b1": list(range(3)),
        }
        res = self._eval(new_t, input)
        print(f"calulate {op},{operands}\n {res}")

    def test_fillna(self):
        pass

    @pytest.mark.parametrize(
        "drop_type,min_freq,features",
        [
            ("first", 0.1, ["a1", "a2", "a3"]),
            ("mode", 0.1, ["a1", "a2", "a3"]),
        ],
    )
    def test_onehot(self, drop_type: str, min_freq: float, features: list):
        # onehot 需要生成相对负责的rule,无法一次性生成expr
        data = {
            "id": [str(i) for i in range(17)],
            "a1": ["K"] + ["F"] * 13 + ["", "M", "N"],
            "a2": [0.1, 0.2, 0.3] * 5 + [0.4] * 2,
            "a3": [1] * 17,
            "y": [0] * 17,
        }
        schema = {"id": "str", "a1": "str", "a2": "float", "a3": "int", "y": "int"}
        t = self._memtable(data, schema=schema)
        row_count = t.count().execute()
        print(f"row_count: {type(row_count)} {row_count}")
        min_freq_count = round(row_count * min_freq)

        def _fit_col(col_name: str):
            col = t[col_name]
            dist_count = col.nunique().execute()

            if dist_count >= 100:
                raise ValueError(
                    f"the feature has too many categories: {col_name}, {dist_count}"
                )

            if drop_type == "mode":
                expr = (
                    t.group_by(col_name)
                    .aggregate(count=col.count())
                    .order_by(ibis.desc("count"))
                    .limit(1)
                )
                res = expr.execute()
                drop_category = res[col_name].iloc[0]
            elif drop_type == "first":
                res = t.limit(1).execute()
                drop_category = res[col_name].iloc[0]
            elif drop_type == "no_drop":
                drop_category = None
            else:
                raise ValueError(f"unsupported drop type: {drop_type}")

            if drop_category is None:
                filtered = t.filter(col.isnull().not_())
            else:
                filtered = t.filter(col != drop_category)
            # 聚合出每个元素个数
            freq_table: pd.DataFrame = (
                filtered.group_by(col_name).aggregate(freq=col.count()).execute()
            )
            # category
            category = []
            infrequent = []
            for _, row in freq_table.iterrows():
                name = row[col_name]
                freq = row["freq"]
                if freq < min_freq_count:
                    infrequent.append(name)
                else:
                    category.append([name])

            if infrequent:
                category.append(infrequent)

            return category, drop_category

        onehot_rules = {}
        drop_rules = {}
        for col_name in features:
            category, drop_category = _fit_col(col_name)
            onehot_rules[col_name] = category
            if drop_category:
                drop_rules[col_name] = drop_category

        # onehot_rules: {'a1': [['F'], ['', 'N', 'M']], 'a2': [[0.3], [0.4], [0.2]], 'a3': []}, drop_rules: {'a1': 'K', 'a2': 0.1, 'a3': 1}
        print(f"onehot_rules: {onehot_rules}, drop_rules: {drop_rules}")
        # apply rule
        exprs = {}
        for col_name, col_rules in onehot_rules.items():
            col = t[col_name]
            for idx, values in enumerate(col_rules):
                new_col = ibis.ifelse(col.isin(values), 1.0, 0.0)
                exprs[f"{col_name}_{idx}"] = new_col
        old_columns = [col for col in t.columns if col not in onehot_rules]
        new_t = t.mutate(**exprs).select(old_columns + list(exprs.keys()))
        res = new_t.execute()
        print(f"onehot res: \n {res}")

    @pytest.mark.parametrize(
        "bin_method,bin_num,features",
        [
            ("eq_range", 3, ["a1", "a2"]),
            ("quantile", 3, ["a1", "a2"]),
            ("", 0, ["s"]),
        ],
    )
    def test_binning(self, bin_method: str, bin_num: int, features: list):
        # 需要对不同列,不同的类型,不同的分箱方法生成不同的规则,也没有办法使用一个表达式完成
        schema = ibis.schema({"a1": "int", "a2": "float", "s": "str"})
        data = {
            "a1": [1, 2, 3, 4, 5],
            "a2": [1.1, 2.2, 3.3, 4.4, 5.5],
            "s": ["s1", "s2", "s2", "s4", "s5"],
        }

        t = self._memtable(data, schema=schema)

        def _fit_feature_bins(col_name: str) -> dict:
            col = t[col_name]
            col_dtype = t.schema().fields[col_name]
            if col_dtype.is_string():
                categories_df = t.drop_null(col_name).distinct(on=col_name).execute()
                categories = categories_df[col_name]
                split_points = sorted(categories)
                return {"type": "str", "split_points": split_points}
            else:
                if bin_method == "quantile":
                    # bins, split_points = pd.qcut(
                    #    f_data, bin_num, labels=False, duplicates='drop', retbins=True
                    # )
                    quantiles = [i / bin_num for i in range(1, bin_num)]
                    split_points_ibis = col.approx_quantile(quantiles)
                    split_points = split_points_ibis.execute()
                    split_points = sorted(set(split_points))
                elif bin_method == "eq_range":
                    # bins, split_points = pd.cut(
                    #     f_data, bin_num, labels=False, duplicates="drop", retbins=True
                    # )
                    # 计算等宽分箱
                    min_val = col.min().execute()
                    max_val = col.max().execute()
                    split_width = (max_val - min_val) / bin_num
                    split_points = [
                        min_val + i * split_width for i in range(1, bin_num)
                    ]
                else:
                    raise ValueError(
                        f"binning_method {self.binning_method} not supported"
                    )

                return {"type": "numeric", "split_points": split_points}

        def _apply_feature_rule(tbl: ibis.Table, rule: dict):
            col_name = rule["name"]
            col = tbl[col_name]

            split_points = rule["split_points"]
            branches = []
            if rule["type"] == "str":
                branches = [(col == c, c) for c in split_points]
            else:
                branches = [(col < c, c) for c in split_points]
                branches.append((col > split_points[-1], split_points[-1]))
            bin_expr = ibis.cases(*branches)

            nt = t.mutate(row_number=ibis.row_number().over())
            nt = nt.select([col_name, "row_number", bin_expr.name("bin_index")])

            na_index_df = nt.filter(col.isnull()).execute()
            na_index = na_index_df["row_number"].to_numpy()

            non_null_nt = nt.filter(col.notnull())
            non_null_df = non_null_nt.execute()

            grouped = non_null_df.groupby("bin_index")["row_number"].apply(list)
            bin_indices = grouped.tolist()
            return bin_indices, na_index

        rules = {}
        for col_name in features:
            rule = _fit_feature_bins(col_name)
            rule["name"] = col_name
            rules[col_name] = rule

        print(f"binning rules: {rules}")
        for col_name, rule in rules.items():
            bin_indices, na_index = _apply_feature_rule(t, rule)
            print(f"use binning rule: {col_name} {bin_indices}, {na_index}")

    def test_woe_binning(self):
        pass

    def test_sql(self):
        schema = {
            "date_id": "date",
            "store_id": "int32",
            "sold_amount": "double",
            "quantity": "double",
        }
        catalog = {"tx": schema}
        sql_str = "SELECT store_id, SUM(sold_amount) AS total FROM tx GROUP BY store_id"
        expr = ibis.parse_sql(sql_str, catalog)

        data = {
            "date_id": pd.to_datetime(["2024-08-01", "2024-08-01", "2024-08-02"]),
            "store_id": [1, 2, 1],
            "sold_amount": [100.0, 200.0, 300.0],
            "quantity": [1, 2, 3],
        }
        res = self._eval(expr, data, table_names=["tx"])
        print(f"sql res:\n {res}")

    def test_score_card(self):
        predict_name: str = "pred"
        predict_score_name: str = "predict_score"
        scaled_value: float = 600
        odd_base: float = 20.0
        pdo: float = 20.0
        positive: int = 1
        min_score: float = 0.0
        max_score: float = 1000.0

        t = ibis.table(schema={"id": "str", "pred": "float"}, name="table")

        factor = pdo / math.log(2)
        offset = scaled_value - factor * math.log(odd_base)
        pred = t[predict_name]
        log_odds = factor * ((pred / (1 - pred)).log())
        if positive == 1:
            score = offset - log_odds
        else:
            score = offset + log_odds

        new_col = ibis.ifelse(
            score <= min_score,
            min_score,
            ibis.ifelse(score >= max_score, max_score, score),
        )
        exprs = {predict_score_name: new_col}
        new_t = t.mutate(**exprs)
        #
        input = {
            "id": ["id1", "id2"],
            "pred": [0.1, 0.2],
        }
        res = self._eval(new_t, input)
        print(f"score_card:\n {res}")
        except_res = [576.959938, 553.561438]
        npt.assert_allclose(res[predict_score_name], except_res)

    @pytest.mark.parametrize(
        "feature,op,bound_value",
        [
            ("a1", "==", "K1"),
        ],
    )
    def test_condition_filter(self, feature, op, bound_value):
        # step 1: compute row mask
        schema = {"id": "str", "a1": "str", "a2": "str", "a3": "int", "a4": "float"}
        t = ibis.table(schema, name="table")
        col_dt = t.schema().fields[feature]
        if col_dt.is_string():
            value_type = str
        elif col_dt.is_floating():
            value_type = float
        elif col_dt.is_integer():
            value_type = int
        else:
            raise ValueError("only support INT, FLOAT, STRING for now")

        value = None
        if op != "NOTNULL":
            if bound_value == "":
                raise ValueError("bound_value is empty")
            bound_value_list = bound_value.split(",")
            values = [value_type(val) for val in bound_value_list]
            value = values if op == "IN" else values[0]

        col = t[feature]
        if op == "==":
            mask_expr = ibis.and_(col.notnull(), col == value)
        elif op == "<":
            mask_expr = ibis.and_(col.notnull(), col < value)
        elif op == "<=":
            mask_expr = ibis.and_(col.notnull(), col <= value)
        elif op == ">":
            mask_expr = ibis.and_(col.notnull(), col > value)
        elif op == ">=":
            mask_expr = ibis.and_(col.notnull(), col >= value)
        elif op == "IN":
            mask_expr = ibis.and_(col.notnull(), col.isin(value))
        elif op == "NOTNULL":
            mask_expr = col.notnull()
        else:
            raise ValueError(f"not support op: {op}")

        new_t = t.mutate(mask=mask_expr).select("mask")

        data = {
            "id": ["1", "2", "3", "4"],
            "a1": ["K5", "K1", None, "K6"],
            "a2": ["A5", "A1", "A2", "A6"],
            "a3": [5, 1, 2, 6],
            "a4": [0, 1, 1, 0],
        }
        res = self._eval(new_t, data)

        # step2: filter by row mask
        # TODO: use ibis?
        df = pd.DataFrame(data)
        filterd_df = df[res["mask"]]
        print(filterd_df)

    @pytest.mark.parametrize("by", [["a"], ["a", "b"]])
    @pytest.mark.parametrize(
        "agg_pairs",
        [
            [("c", "sum")],
            [("c", "count"), ("d", "sum")],
        ],
    )
    def test_stat_groupby(self, by: list[str], agg_pairs: list[tuple[str, str]]):
        data = {
            "a": ["9", "6", "5", "5"],
            "b": [5, 5, 6, 7],
            "c": [1, 1, 2, 4],
            "d": [11, 55, 1, 99],
        }
        schema = {"a": "str", "b": "int", "c": "float", "d": "int"}
        t = ibis.table(schema=schema, name="table")

        agg_map = {
            "count": lambda x: x.count(),
            "sum": lambda x: x.sum(),
            "mean": lambda x: x.mean(),
            "min": lambda x: x.min(),
            "max": lambda x: x.max(),
            "var": lambda x: x.var(),  # Variance
        }

        grouped = t.group_by(by)
        aggs = []
        for p in agg_pairs:
            agg_col = p[0]
            agg_fn = p[1]
            if agg_fn not in agg_map:
                raise ValueError(f"Unsupported aggregation function: {agg_fn}")
            expr = agg_map[agg_fn](t[agg_col]).name(f"{agg_col}_{agg_fn}")
            aggs.append(expr)

        new_t = grouped.aggregate(aggs)
        res = self._eval(new_t, data)
        print(f"groupby: {by}\n {res}")

    def test_stat_pearsonr(self):
        pass

    @pytest.mark.parametrize("features", [["a1", "a2"]])
    def test_stat_vif(self, features: list[str]):
        pass

    def test_stat_psi(self):
        pass

    @pytest.mark.parametrize("features", [["a", "b", "c", "d", "f"]])
    def test_stat_table_stats(self, features: list[str]):
        # duckdb has impl skewness, register for ibis.
        @ibis.udf.agg.builtin
        def skewness(a: float) -> float:
            """Compute the skewness."""

        schema_pairs = {"a": "int", "b": "int", "c": "int", "d": "int", "f": "str"}
        table = ibis.table(schema=schema_pairs, name="table")

        def _compute_if(col_expr: ibis.Column, fn):
            if not col_expr.type().is_string():
                return fn(col_expr).cast("float64")
            return ibis.null("float64")

        def _central_moment(col: ibis.Column, k):
            #  "t0"."c" - AVG("t0"."c") OVER (ROWS BETWEEN UNBOUNDED PRECEDING AND...
            mean_expr = col.mean()
            return ((col - mean_expr) ** k).mean()

        def get_stat_expr(col_expr: ibis.Column, order: int):
            return ibis.struct({
                "order": ibis.literal(order),
                "column": ibis.literal(col_expr.get_name()),
                "datatype": ibis.literal(str(col_expr.type())),
                "nunique": col_expr.nunique(),
                "count": col_expr.count(),
                "count_na": col_expr.isnull().sum(),
                "na_ratio": col_expr.isnull().sum() / col_expr.count(),
                "min": _compute_if(col_expr, lambda c: c.min()),
                "max": _compute_if(col_expr, lambda c: c.max()),
                "mean": _compute_if(col_expr, lambda c: c.mean()),
                "var": _compute_if(col_expr, lambda c: c.var()),
                "std": _compute_if(col_expr, lambda c: c.std()),
                "sem": _compute_if(col_expr, lambda c: c.std() / c.count().sqrt()),
                "skew": _compute_if(col_expr, lambda c: skewness(c)),
                "kurtosis": _compute_if(col_expr, lambda c: c.kurtosis()),
                "q1": _compute_if(col_expr, lambda c: c.quantile(0.25)),
                "q2": _compute_if(col_expr, lambda c: c.quantile(0.5)),
                "q3": _compute_if(col_expr, lambda c: c.quantile(0.75)),
                "moment_2": _compute_if(col_expr, lambda c: (c**2).mean()),
                "moment_3": _compute_if(col_expr, lambda c: (c**3).mean()),
                "moment_4": _compute_if(col_expr, lambda c: (c**4).mean()),
                # "central_moment_2": _compute_if(
                #     col_expr, lambda c: _central_moment(c, 2)
                # ),
                # "central_moment_3": _compute_if(
                #     col_expr, lambda c: _central_moment(c, 3)
                # ),
                # "central_moment_4": _compute_if(
                #     col_expr, lambda c: _central_moment(c, 4)
                # ),
                "sum": _compute_if(col_expr, lambda c: c.sum()),
                "sum_2": _compute_if(col_expr, lambda c: (c**2).sum()),
                "sum_3": _compute_if(col_expr, lambda c: (c**3).sum()),
                "sum_4": _compute_if(col_expr, lambda c: (c**4).sum()),
            }).name("stats")

        stats_tables = []
        for idx, col_name in enumerate(features):
            col = table[col_name]
            stat_expr = get_stat_expr(col, idx)
            stats_table = table.aggregate([stat_expr])
            stats_tables.append(stats_table)

        result_expr = ibis.union(*stats_tables)
        struct_fields = result_expr["stats"].type().names
        mutations = {field: result_expr["stats"][field] for field in struct_fields}

        result_expr = (
            result_expr.mutate(**mutations)
            .drop("stats")
            .order_by("order")
            .drop("order")
        )

        data = {
            "a": [9, 6, 5, 5],
            "b": [5, 5, 6, 7],
            "c": [1, 1, 2, 4],
            "d": [11, 55, 1, 99],
            "f": ["hello", "world", "are you", "ok?"],
        }
        res = self._eval(result_expr, data)
        print(f"table_stats: {features}\n {res}")

    def _memtable(self, data: dict, schema, name: str = "table") -> ibis.Table:
        return ibis.memtable(data, schema=schema, name=name)

    def _eval(
        self, expr: ibis.Table, *inputs: dict, table_names: list[str] | None = None
    ) -> pd.DataFrame:
        if table_names is None:
            table_names = ["table"]

        if _ibis_engine == "duckdb":
            return self._eval_duckdb(expr, inputs, table_names)
        else:
            return self._eval_substrait(expr, inputs, table_names)

    def _eval_duckdb(
        self,
        expr: ibis.Table,
        inputs: list[dict],
        table_names: list[str] | None = None,
        udfs: list | None = None,
    ):
        if not table_names:
            table_names = ["table"]
        assert len(inputs) == len(table_names)

        sql = ibis.to_sql(expr, dialect="duckdb")

        print(f"eval sql: \n{sql}\n")

        import duckdb
        import pandas as pd

        conn = duckdb.connect(":memory:")
        for data, name in zip(inputs, table_names, strict=True):
            df = pd.DataFrame(data)
            conn.register(name, df)

        if udfs:
            for fn in udfs:
                name = fn.__name__
                conn.create_function(name, fn)

        result_df = conn.execute(sql).fetchdf()
        return result_df

    def _eval_substrait(
        self, expr: ibis.Table, inputs: list[dict], table_names: list[str] | None = None
    ):
        # You need to run cmd 'pip install ibis-substrait' manually.
        # It will cause the ibis-framework version to roll back
        import pyarrow as pa
        import pyarrow.substrait as substrait  # noqa: F401
        from ibis_substrait.compiler.core import SubstraitCompiler

        assert len(inputs) == len(table_names)

        # encode ir
        compiler = SubstraitCompiler()
        plan = compiler.compile(expr)
        fn_txt = plan.SerializeToString()

        # # decode ir
        # substrait_plan = substrait.Plan()
        # substrait_plan.ParseFromString(fn_txt)

        # exec ir use pyarrow

        tables = {
            name: pa.Table.from_pydict(data)
            for name, data in zip(table_names, inputs, strict=True)
        }

        def table_provider(names, schema):
            if not names:
                raise Exception("No names provided")
            for name in names:
                if name in tables:
                    return tables[name]
            raise Exception(f"Unrecognized table name, {names}, {table_names}")

        reader = pa.substrait.run_query(fn_txt, table_provider=table_provider)
        result = reader.read_all()
        return result



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration: global symbols read -> compute -> write new global symbol.

Flow:
 1. Spawn two HTTP runtimes via the shared fixture infrastructure.
 2. Use the client SDK to upload ``x`` only to party P0's global symbol table.
 3. Evaluate a function that reads ``x`` on P0, adds 1 on P0, sends the result to P1, and writes ``y`` from P1.
 4. Confirm the computation result is now held by P1 only.
 5. Use the client SDK to fetch ``y`` from the P1 runtime and verify ``y == x + 1``.

This tests end-to-end interaction between:
 - Global symbol CRUD endpoints
 - ``basic.read`` path using ``symbols://`` scheme
 - Driver evaluation and fetch
 - HttpExecutorClient helpers for symbol CRUD
"""

from __future__ import annotations

import asyncio

import numpy as np
import pytest

import mplang
import mplang.simp as simp
from mplang.core.cluster import ClusterSpec, Device, Node, RuntimeInfo
from mplang.core.dtype import DType
from mplang.core.tensor import TensorType
from mplang.kernels.value import TensorValue
from mplang.ops import basic as fb
from mplang.runtime.client import HttpExecutorClient

pytest_plugins = ("tests.utils.server_fixtures",)


@pytest.mark.parametrize("http_servers", [2], indirect=True)
def test_global_symbol_roundtrip(http_servers):
    node_ids = ["P0", "P1"]
    node_addrs = dict(zip(node_ids, http_servers.addresses, strict=True))

    # Cluster spec
    nodes = {}
    for nid, addr in node_addrs.items():
        r = int(nid[1:])
        nodes[f"node{r}"] = Node(
            name=f"node{r}",
            rank=r,
            endpoint=addr,
            runtime_info=RuntimeInfo(version="test", platform="test", op_bindings={}),
        )
    local_devices = {
        f"local_{n.rank}": Device(name=f"local_{n.rank}", kind="local", members=[n])
        for n in nodes.values()
    }
    spu_device = Device(
        name="SPU_0",
        kind="SPU",
        members=list(nodes.values()),
        config={"protocol": "SEMI2K", "field": "FM64"},
    )
    spec = ClusterSpec(nodes=nodes, devices={**local_devices, "SPU_0": spu_device})
    driver = mplang.Driver(spec)

    arr = np.arange(5, dtype=np.int32)
    tensor_meta = {"tensor": {"dtype": "I32", "shape": list(arr.shape)}}

    async def upload_symbol(addr: str, symbol_name: str, data: np.ndarray) -> None:
        client = HttpExecutorClient(addr)
        try:
            # Wrap numpy array in TensorValue before sending
            await client.create_global_symbol(
                symbol_name, TensorValue(data), tensor_meta
            )
        finally:
            await client.close()

    # Seed only P0 with the initial value.
    asyncio.run(upload_symbol(http_servers.addresses[0], "x", arr))

    @mplang.function
    def compute_chain():
        ty = TensorType(DType.from_any(np.int32), (5,))
        # P0 reads x from its local global symbol table.
        x_p0 = simp.runAt(0, fb.read)(path="symbols://x", ty=ty)

        # P0 computes x + 1.
        def add_one(a):
            import jax.numpy as jnp

            return jnp.add(a, jnp.ones_like(a))

        x_inc_p0 = simp.runAt(0, add_one)(x_p0)

        # Transfer the result to P1 and write it out as symbols://y from P1.
        x_p1 = simp.p2p(0, 1, x_inc_p0)
        written = simp.runAt(1, fb.write)(x_p1, path="symbols://y")

        return written

    out = mplang.evaluate(driver, compute_chain)
    fetched = mplang.fetch(driver, out)

    # Only P1 should now hold the materialized value.
    assert fetched[0] is None, "P0 should not retain the result after p2p transfer"
    np.testing.assert_array_equal(fetched[1], arr + 1)

    async def fetch_y_from_p1() -> np.ndarray:
        client = HttpExecutorClient(http_servers.addresses[1])
        try:
            result = await client.get_global_symbol("y")
            return np.asarray(result)
        finally:
            await client.close()

    y_val = asyncio.run(fetch_y_from_p1())
    np.testing.assert_array_equal(y_val, arr + 1)



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Core components for multi-party computation.

This package provides the fundamental building blocks for multi-party computation,
including type systems, tracing mechanisms, and interpreter contexts.
"""

# Core type system
# Communication interfaces & core symbols
from mplang.core.comm import (
    CollectiveMixin,
    CommunicatorBase,
    ICollective,
    ICommunicator,
)
from mplang.core.dtype import DType
from mplang.core.interp import InterpContext, InterpVar
from mplang.core.mask import Mask
from mplang.core.mpobject import MPContext, MPObject
from mplang.core.mptype import MPType, Rank, Shape
from mplang.core.pfunc import PFunction
from mplang.core.primitive import (
    constant,
    debug_print,
    function,
    pconv,
    peval,
    prand,
    prank,
    pshfl,
    pshfl_s,
    psize,
    set_mask,
    uniform_cond,
    while_loop,
)
from mplang.core.table import TableLike, TableType
from mplang.core.tensor import TensorLike, TensorType
from mplang.core.tracer import TraceContext, TracedFunction, TraceVar, VarNamer, trace

__all__ = [
    "CollectiveMixin",
    "CommunicatorBase",
    "DType",
    "ICollective",
    "ICommunicator",
    "InterpContext",
    "InterpVar",
    "MPContext",
    "MPObject",
    "MPType",
    "Mask",
    "PFunction",
    "Rank",
    "Shape",
    "TableLike",
    "TableType",
    "TensorLike",
    "TensorType",
    "TraceContext",
    "TraceVar",
    "TracedFunction",
    "VarNamer",
    "constant",
    "debug_print",
    "function",
    "pconv",
    "peval",
    "prand",
    "prank",
    "pshfl",
    "pshfl_s",
    "psize",
    "set_mask",
    "trace",
    "uniform_cond",
    "while_loop",
]



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from urllib.parse import ParseResult, urlparse

import numpy as np
import pandas as pd

from mplang.core.table import TableType
from mplang.core.tensor import TensorType
from mplang.kernels.base import KernelContext
from mplang.utils import table_utils


@dataclass(frozen=True)
class ResolvedURI:
    """Result of resolving a resource path into a normalized form.

    Attributes:
      scheme: The URI scheme (e.g., 'file', 's3', 'mem', 'var', 'secret').
      raw: The original path string as provided by the user.
      parsed: The ParseResult if a scheme was present; otherwise None.
      local_path: For file paths: concrete filesystem path (absolute or as given).
    """

    scheme: str
    raw: str
    parsed: ParseResult | None
    local_path: str | None


def resolve_uri(path: str) -> ResolvedURI:
    """Resolve a user-provided resource location into a normalized URI form.

    This helper accepts plain filesystem paths and RFC 3986 style URIs. A path
    is treated as ``file`` when ``urlparse(path).scheme`` is empty. Detection
    no longer depends on the presence of the literal substring ``"://"`` so
    that forms like ``mem:foo`` (no slashes) are still recognized as a URI.

    Captured fields
    - ``scheme``: Lower-cased scheme (``file`` when absent)
    - ``raw``: Original input
    - ``parsed``: ``ParseResult`` when a scheme was provided, else ``None``
    - ``local_path``: Filesystem path for ``file`` scheme, else ``None``

    Supported (pluggable) schemes out-of-the-box:
      * ``file`` (default)
      * ``mem``
      * ``s3`` (stub)
      * ``secret`` (stub)
      * ``symbols`` (registered server-side)

    Examples
    >>> resolve_uri("data/train.npy").scheme
    'file'
    >>> resolve_uri("mem:dataset1").scheme
    'mem'
    >>> resolve_uri("mem://dataset1").scheme  # both forms acceptable
    'mem'
    >>> resolve_uri("symbols://shared_model").scheme
    'symbols'
    >>> resolve_uri("file:///tmp/x.npy").local_path
    '/tmp/x.npy'
    """

    pr = urlparse(path)
    if not pr.scheme:
        return ResolvedURI("file", path, None, path)

    scheme = pr.scheme.lower()
    local_path: str | None = None
    if scheme == "file":
        local_path = pr.path
        if pr.netloc and not local_path.startswith("/"):
            local_path = f"//{pr.netloc}/{pr.path}"
    return ResolvedURI(scheme, path, pr, local_path)


class DataProvider:
    """Abstract base for data providers.

    Minimal contract: read/write by URI and type spec. Providers may ignore the
    type spec but SHOULD validate when feasible.
    """

    def read(
        self, uri: ResolvedURI, out_spec: TensorType | TableType, *, ctx: KernelContext
    ) -> Any:
        raise NotImplementedError

    def write(self, uri: ResolvedURI, value: Any, *, ctx: KernelContext) -> None:
        raise NotImplementedError


_REGISTRY: dict[str, DataProvider] = {}


def register_provider(
    scheme: str, provider: DataProvider, *, replace: bool = False, quiet: bool = False
) -> None:
    """Register a provider implementation.

    Args:
        scheme: URI scheme handled (case-insensitive)
        provider: Implementation
        replace: If False and scheme exists -> ValueError
        quiet: If True, suppress duplicate log messages when replacing
    """
    import logging

    key = scheme.lower()
    if not replace and key in _REGISTRY:
        raise ValueError(f"provider already registered for scheme: {scheme}")
    if replace and key in _REGISTRY and not quiet:
        logging.info(f"Replacing existing provider for scheme '{scheme}'")
    _REGISTRY[key] = provider


def get_provider(scheme: str) -> DataProvider | None:
    return _REGISTRY.get(scheme.lower())


# ---------------- Default Providers ----------------


class FileProvider(DataProvider):
    """Local filesystem provider.

    For tables: CSV bytes via table_utils.
    For tensors: NumPy .npy via np.load/np.save.
    """

    def read(
        self, uri: ResolvedURI, out_spec: TensorType | TableType, *, ctx: KernelContext
    ) -> Any:
        path = uri.local_path or uri.raw
        if isinstance(out_spec, TableType):
            with open(path, "rb") as f:
                csv_bytes = f.read()
            return table_utils.csv_to_dataframe(csv_bytes)
        # tensor path
        return np.load(path)

    def write(self, uri: ResolvedURI, value: Any, *, ctx: KernelContext) -> None:
        import os

        path = uri.local_path or uri.raw
        dir_name = os.path.dirname(path)
        if dir_name:
            os.makedirs(dir_name, exist_ok=True)
        # Table-like to CSV bytes
        if hasattr(value, "__dataframe__") or isinstance(value, pd.DataFrame):
            csv_bytes = table_utils.dataframe_to_csv(value)  # type: ignore
            with open(path, "wb") as f:
                f.write(csv_bytes)
            return
        # Tensor-like via numpy
        np.save(path, np.asarray(value))


class MemProvider(DataProvider):
    """In-memory per-runtime KV provider (per rank, per session/runtime)."""

    STATE_KEY = "resource.providers.mem"

    @staticmethod
    def _store(ctx: KernelContext) -> dict[str, Any]:
        # Use ensure_state so creation is atomic & centralized; enforce dict.
        store = ctx.runtime.ensure_state(MemProvider.STATE_KEY, dict)
        if not isinstance(store, dict):  # pragma: no cover - defensive
            raise TypeError(
                f"runtime state key '{MemProvider.STATE_KEY}' expected dict, got {type(store).__name__}"
            )
        return store  # type: ignore[return-value]

    def read(
        self, uri: ResolvedURI, out_spec: TensorType | TableType, *, ctx: KernelContext
    ) -> Any:
        store = self._store(ctx)
        key = uri.raw
        if key not in store:
            raise FileNotFoundError(f"mem resource not found: {key}")
        return store[key]

    def write(self, uri: ResolvedURI, value: Any, *, ctx: KernelContext) -> None:
        store = self._store(ctx)
        store[uri.raw] = value


class S3Provider(DataProvider):
    """Placeholder S3 provider. Install external plugin to enable."""

    def read(
        self, uri: ResolvedURI, out_spec: TensorType | TableType, *, ctx: KernelContext
    ) -> Any:
        raise NotImplementedError(
            "S3 provider not installed. Provide an external plugin via register_provider('s3', ...) ."
        )

    def write(self, uri: ResolvedURI, value: Any, *, ctx: KernelContext) -> None:
        raise NotImplementedError(
            "S3 provider not installed. Provide an external plugin via register_provider('s3', ...) ."
        )


class SecretProvider(DataProvider):
    """Placeholder secret provider. Integrate with KMS/secret manager via plugin."""

    def read(
        self, uri: ResolvedURI, out_spec: TensorType | TableType, *, ctx: KernelContext
    ) -> Any:
        raise NotImplementedError(
            "secret provider not installed. Provide an external plugin via register_provider('secret', ...) ."
        )

    def write(self, uri: ResolvedURI, value: Any, *, ctx: KernelContext) -> None:
        raise NotImplementedError(
            "secret provider not installed. Provide an external plugin via register_provider('secret', ...) ."
        )


# Register default providers
register_provider("file", FileProvider())
register_provider("mem", MemProvider())
# Stubs to signal missing providers explicitly (can be overridden by plugins)
register_provider("s3", S3Provider())
register_provider("secret", SecretProvider())
